##################################################################################
# This file is part of 21cmCAST.
#
# Copyright (c) 2023, Gaétan Facchinetti
#
# 21cmCAST is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. 21cmCAST is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with 21cmCAST. 
# If not, see <https://www.gnu.org/licenses/>.
#
# -------------------------------------------------------------------------------
#
# Parts of this code have been copied and modified 
# from https://github.com/charlottenosam/21cmfish
# 
# - MIT License
# -
# - Copyright (c) 2019, Charlotte Mason
# - 
# - Permission is hereby granted, free of charge, to any person obtaining a copy
# - of this software and associated documentation files (the "Software"), to deal
# - in the Software without restriction, including without limitation the rights
# - to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# - copies of the Software, and to permit persons to whom the Software is
# - furnished to do so, subject to the following conditions:
# - 
# - The above copyright notice and this permission notice shall be included in all
# - copies or substantial portions of the Software.
# - 
# - THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# - IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# - FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# - AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# - LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# - OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# - SOFTWARE.
##################################################################################

import configparser
from copy import deepcopy
import os

from py21cmcast import tools as p21c_tools
import warnings
import numpy as np
import pickle

PY21CMFAST = True
try:
    import py21cmfast as p21f
except ImportError:
    PY21CMFAST = False
    p21f = None
    warnings.warn("21cmFAST could not be found! 21cmCAST can work on cached data but will not be able to process any new lightcones.")



def init_runs(config_file: str, q_scale: float = 3., clean_existing_dir: bool = False, verbose = False) -> None :

    """ 
    Initialise the runs for a fisher analysis according to 
    a fiducial model defined in config_file
    
    Params : 
    --------
    config_file : str
        Path to the config file representing the fiducial
    q_scale : float, optional
        Gives the points where to compute the derivative in pourcentage of the fiducial parameters
    erase_dir : bool, optional
        If True forces the creation of a folder 
    """

    config = configparser.ConfigParser(delimiters=':')
    config.optionxform = str

    config.read(config_file)

    name            = config.get('run', 'name')
    output_dir      = config.get('run', 'output_dir')
    cache_dir       = config.get('run', 'cache_dir')

    extra_params = {}

    try:
        lightcone_q       = p21c_tools.read_config_params(config.items('lightcone_quantities'))
    except configparser.NoSectionError:
        if verbose: 
            warnings.warn("No section lightcone_quantities provided")
        lightcone_q = {}

    try:
        global_q       = p21c_tools.read_config_params(config.items('global_quantities'))
    except configparser.NoSectionError:
        if verbose: 
            warnings.warn("No section global_quantities provided")
        global_q = {}    

    
    extra_params      = p21c_tools.read_config_params(config.items('extra_params'), int_type = False)
    user_params       = p21c_tools.read_config_params(config.items('user_params'))
    flag_options      = p21c_tools.read_config_params(config.items('flag_options'))
    astro_params_fid  = p21c_tools.read_config_params(config.items('astro_params'), int_type = False, allow_lists=False)
    cosmo_params_fid  = p21c_tools.read_config_params(config.items('cosmo_params'), int_type = False, allow_lists=False)


    try: 
        astro_params_vary = p21c_tools.read_config_params(config.items('astro_params_vary'), int_type = False, allow_lists=True)
    except configparser.NoSectionError:
        if verbose: 
            warnings.warn("No section astro_params_vary provided")
        astro_params_vary = {}

    try: 
        cosmo_params_vary = p21c_tools.read_config_params(config.items('cosmo_params_vary'), int_type = False, allow_lists=True)
    except configparser.NoSectionError:
        if verbose: 
            warnings.warn("No section cosmo_params_vary provided")
        astro_params_vary = {}

    try:
        one_side_der      = p21c_tools.read_config_params(config.items('one_side_param_derivative'))
    except configparser.NoSectionError:
        if verbose: 
            warnings.warn("No section one_side_param_derivative provided")
        one_side_der = {}

    try: 
        percentage        = p21c_tools.read_config_params(config.items('percentage'))
    except configparser.NoSectionError:
        if verbose: 
            warnings.warn("No section percentage provided")
        percentage = {}


    # Make the directory to store the outputs and everything
    output_run_dir = output_dir + "/" + name.upper() + "/"
    existing_dir = p21c_tools.make_directory(output_run_dir, clean_existing_dir = clean_existing_dir)

    if existing_dir is True:
        print('WARNING: Cannot create a clean new folder because clean_existing_dir is False')
        return 

    astro_params_run_all = {}
    astro_params_run_all['FIDUCIAL'] = astro_params_fid

    cosmo_params_run_all = {}
    cosmo_params_run_all['FIDUCIAL'] = cosmo_params_fid

    params_vary  = {}
    class_params = {}
    
    for param, value in astro_params_vary.items():
        params_vary[param] = value
        class_params[param] = "ASTRO"
    for param, value in cosmo_params_vary.items():
        params_vary[param] = value
        class_params[param] = "COSMO"

    for param, values in params_vary.items(): 
        
        for value in values:

            _side_der = one_side_der.get(param, None)
            _percentage = percentage.get(param, True)
            
            if _side_der is None :
                vary_array = np.array([-100., 100.])
            else:
                if _side_der == '+': 
                    vary_array = np.array([+100.])
                elif _side_der == '-':
                    vary_array = np.array([-100.])
                else: 
                    ValueError("The arguments of one_side_param_derivative have to be '+' or '-'")
            try:
                if class_params[param] == "ASTRO":
                    p_fid = astro_params_fid[param] 
                elif class_params[param] == "COSMO":
                    p_fid = cosmo_params_fid[param]
                else:
                    ValueError("Parameter varied is neither astro or cosmo.")
            except KeyError:
                warnings.warn("Cannon treat parameter:", param, " as no fiducial value was passed")
                    
        
            if isinstance(value, float) and value > 0:
                q = value/vary_array
            else : 
                q = q_scale/vary_array

            if _percentage is False:
                p = 100.*q
            else:
                p = p_fid*(1+q)
                print('Parameter ' + str(param) + ' varied by ' + str(value) + ' percent of the fiducial')

            astro_params_run = astro_params_fid.copy()
            cosmo_params_run = cosmo_params_fid.copy()

            for i, pp in enumerate(p):
                if class_params[param] == "ASTRO":
                    astro_params_run[param] = pp
                    astro_params_run_all[param + '_{:.4e}'.format(pp)] = astro_params_run.copy()
                elif class_params[param] == "COSMO":
                    cosmo_params_run[param] = pp
                    cosmo_params_run_all[param + '_{:.4e}'.format(pp)] = cosmo_params_run.copy()
                
    # Write down the separate config files
    irun = 0
    for key, astro_params in astro_params_run_all.items() : 
        p21c_tools.write_config_params(output_run_dir + '/Config_' + key + ".config", name, output_run_dir, cache_dir, 
                                       lightcone_q, global_q, extra_params, user_params, flag_options, astro_params, cosmo_params_fid, key)
        irun = irun + 1
        
    for key, cosmo_params in cosmo_params_run_all.items() : 
        p21c_tools.write_config_params(output_run_dir + '/Config_' + key + ".config", name, output_run_dir, cache_dir, 
                                       lightcone_q, global_q, extra_params, user_params, flag_options, astro_params_fid, cosmo_params, key)
        irun = irun + 1
    
    return 


def init_grid_runs(config_file: str, clean_existing_dir: bool = False, verbose:bool = False) -> None:
    
    config = configparser.ConfigParser(delimiters=':')
    config.optionxform = str

    config.read(config_file)

    name            = config.get('run', 'name')
    output_dir      = config.get('run', 'output_dir')
    cache_dir       = config.get('run', 'cache_dir')

    output_run_dir = output_dir + "/" + name.upper() + "/"
    existing_dir = p21c_tools.make_directory(output_run_dir, clean_existing_dir = clean_existing_dir)

    if existing_dir is True:
        print('WARNING: Cannot create a clean new folder because clean_existing_dir is False')
        return 

    try:
        lightcone_q       = p21c_tools.read_config_params(config.items('lightcone_quantities'))
    except configparser.NoSectionError:
        if verbose: 
            warnings.warn("No section lightcone_quantities provided")
        lightcone_q = {}

    try:
        global_q       = p21c_tools.read_config_params(config.items('global_quantities'))
    except configparser.NoSectionError:
        if verbose: 
            warnings.warn("No section global_quantities provided")
        global_q = {}    



    extra_params = p21c_tools.read_config_params(config.items('extra_params'), int_type = False)
    user_params  = p21c_tools.read_config_params(config.items('user_params'))
    flag_options = p21c_tools.read_config_params(config.items('flag_options'))
    astro_params = p21c_tools.read_config_params(config.items('astro_params'), int_type = False, allow_lists=True)
    cosmo_params = p21c_tools.read_config_params(config.items('cosmo_params'), int_type = False, allow_lists=True)


    params_array, params_single = collect_params_arrays(astro_params | cosmo_params)

    all_indices = set()
    params = []
    values = []

    # Goes over all astro and cosmo params and makes set of all possible index combinations to get all possible values
    for param_key, param_values in (astro_params | cosmo_params).items():
        all_indices = indices_combinations(param_values, index_combinations=all_indices)
        params.append(param_key)
        values.append(param_values)


    # Create and write the database in a human readable format
    with open(output_run_dir + "/database.txt", 'a') as file:
        
        print("# DATAFILE", file = file)
        print("##############################", file = file)
        print("# Unvaried astro parameters", file = file)

        for param in params_single:
            if param in astro_params:
                print(param + ' : ' + str(astro_params[param][0]) + ' | ', file = file, end ='')
        
        print("", file = file)
        print("# Unvaried cosmo parameters", file = file)
        for param in params_single:
            if param in cosmo_params:
                print(param + ' : ' + str(cosmo_params[param][0]) + ' | ', file = file, end ='')

        print("", file = file)

        print("##############################", file = file)
        print("# Varied parameters", file = file) 
        
        print("# Index :", file = file, end = '') 

        for param in params_array:
            print(' | ' + param, file = file, end ='')

        print("", file = file)

        database = {param : [] for param in (astro_params | cosmo_params)}

        for global_ind, indexes in enumerate(all_indices):
            astro_params_print = {}
            cosmo_params_print = {}
            for ival, i in enumerate(indexes):
                if params[ival] in astro_params:
                    astro_params_print = astro_params_print | {params[ival] :  values[ival][i]}
                if params[ival] in cosmo_params:
                    cosmo_params_print = cosmo_params_print | {params[ival] :  values[ival][i]}
            
                database[params[ival]].append(values[ival][i])
            
            p21c_tools.write_config_params(output_run_dir + '/Config_' + str(global_ind) + ".config", name, output_run_dir, cache_dir, 
                                    lightcone_q, global_q, extra_params, user_params, flag_options, astro_params_print, cosmo_params_print, str(global_ind))

            # ---------------------------- #
            for param in params_array:
                if param in astro_params:
                    print(' ' + str(astro_params_print[param]), file=file, end='')
                if param in cosmo_params:
                    print(' ' + str(cosmo_params_print[param]), file=file, end='')
            print("", file=file) 
            # ---------------------------- #
                
        file.close()

    # Write database in a machine readable format
    with open(output_run_dir + "/database.pkl", 'wb') as file:
        pickle.dump(database, file)
    
    print(str(len(all_indices)) + ' config files generated')


def read_grid_database(dir_path:str):

    with open(dir_path + '/database.pkl', 'rb') as file:
        database = pickle.load(file)

    return database


def find_in_grid_database(database, **kwargs):

    indices = []

    for key, value in kwargs.items():
        indices.append([i for i, j in enumerate(database[key]) if j == value])


    common_indices = set(indices[0])  # Initialize with the first list
    for sublist in indices[1:]:
        common_indices = common_indices.intersection(sublist)

    # Convert the set to a list if needed
    return list(common_indices)



def collect_params_arrays(params):

    params_array = []
    params_single = []

    for key, value in params.items():
        if len(value) > 1:
            params_array.append(key)
        else:
            params_single.append(key)
            
    return params_array, params_single


def indices_combinations(*arrays, index_combinations=None):    
    # Initialize index_combinations as an empty set for the first call
    if index_combinations is None:
        index_combinations = set()

    # Base case: if there are no more arrays, return the current set of index combinations
    if not arrays:
        return index_combinations

    # Extract the first array from the arguments
    current_array = arrays[0]

    # If this is the first array, initialize the index_combinations with its indices
    if not index_combinations:
        for i in range(len(current_array)):
            index_combinations.add((i,))
    else:
        # Generate new combinations by combining each existing combination with indices of the current array
        new_combinations = set()
        for index_combination in index_combinations:
            for i in range(len(current_array)):
                new_combinations.add(index_combination + (i,))
        index_combinations = new_combinations

    # Recur with remaining arrays and the updated index_combinations
    return indices_combinations(*arrays[1:], index_combinations=index_combinations)
       

    




def make_config_one_varying_param(config_file: str, param_name: str, values, **kwargs):

    config = configparser.ConfigParser(delimiters=':')
    config.optionxform = str

    config.read(config_file)

    name            = config.get('run', 'name')
    output_dir      = config.get('run', 'output_dir')
    cache_dir       = config.get('run', 'cache_dir')

    extra_params = {}
    
    try: 
        extra_params['min_redshift']  = float(config.get('extra_params','min_redshift'))
    except configparser.NoOptionError: 
        print("Warning: min_redshift set to 5 by default")
        extra_params['min_redshift']  = 5

    try:
        extra_params['max_redshift']    = float(config.get('extra_params','max_redshift'))
    except configparser.NoOptionError:
        print("Warning: max_redshift set to 35 by default")
        extra_params['max_redshift']  = 35   

    try:
        extra_params['coarsen_factor']  = int(config.get('extra_params', 'coarsen_factor'))
    except configparser.NoOptionError:
        print("Warning: coarsen factor undifined")

    user_params       = p21c_tools.read_config_params(config.items('user_params'))
    flag_options      = p21c_tools.read_config_params(config.items('flag_options'))
    astro_params_fid  = p21c_tools.read_config_params(config.items('astro_params'), int_type = False)
    
    # Make the directory to store the outputs and everything
    output_run_dir = output_dir + "/" + name.upper() + "/"

    mod_astro_params = kwargs.get('mod_astro_params', None)
    mod_flag_options = kwargs.get('mod_flag_options', None)
    add_file_name    = kwargs.get('add_file_name', '')

    if add_file_name != '':
        add_file_name = add_file_name + '_' 

    if mod_astro_params is not None:
        for param in mod_astro_params.keys():
            if not param in astro_params_fid:
                raise KeyError('This parameter is not primarly defined in the config file')
            astro_params_fid[param] = mod_astro_params[param]

    if mod_flag_options is not None:
        for flag in mod_flag_options.keys():
            if not flag in flag_options:
                raise KeyError('This flag_option is not primarly defined in the config file')
            flag_options[flag] = mod_flag_options[flag]

    astro_params = astro_params_fid

    if not param_name in astro_params_fid:
        raise KeyError('This parameter is not primarly defined in the config file')
    
    for ival, val in enumerate(values):
        astro_params[param_name] = val
        filename = output_run_dir + '/Config_' + param_name + '_' + add_file_name + '{:.4e}'.format(val) + ".config"
        key      = param_name + '_' + add_file_name + '{:.4e}'.format(val)
        #print(filename, astro_params, flag_options, key)
        p21c_tools.write_config_params(filename, name, cache_dir, extra_params, user_params, flag_options, astro_params, key)





def run_lightcone_from_config(config_file: str, n_omp: int = None, random_seed: int = None, **kwargs) :

    """ 
    ## Run a lightcone from a config file 

    Parameters
    ----------
    config_file: str
        path of the configuration file
    n_omp: int
        number of threads to use
    run_id : str, optional
        id of the run we are currently looking at
    random_seed : int
        random_seed with which to run 21cmFAST

    Returns
    ----------
    lightcone: Lightcone object (see 21cmFAST)
        lightcone of the run
    run_id: string
        identifier of the run
    """

    ####################### Getting the data ############################

    config = configparser.ConfigParser(delimiters=':')
    config.optionxform = str
    config.read(config_file)

    name            = config.get('run', 'name')
    run_id          = config.get('run', 'run_id', fallback='') 
    output_dir      = config.get('run', 'output_dir')
    cache_dir       = config.get('run', 'cache_dir')
    

    ## Need to implement this next time
    lightcone_q     = p21c_tools.read_config_params(config.items('lightcone_quantities'))
    global_q        = p21c_tools.read_config_params(config.items('global_quantities'))

    extra_params    = p21c_tools.read_config_params(config.items('extra_params'), int_type=False)
    user_params     = p21c_tools.read_config_params(config.items('user_params'))
    flag_options    = p21c_tools.read_config_params(config.items('flag_options'))
    astro_params    = p21c_tools.read_config_params(config.items('astro_params'), int_type=False)
    cosmo_params    = p21c_tools.read_config_params(config.items('cosmo_params'), int_type=False)

    # manually set the number of threads
    if n_omp is not None: 
        user_params['N_THREADS'] = int(n_omp)

    seed_str : str = ''
    if random_seed is not None:
        seed_str = 'rs' + str(random_seed) + '_'

    cache_path = cache_dir + name.upper() + '/run_' + seed_str + run_id + '/'
    
    ####################### Running the lightcone ############################

    lightcone_quantities = ()
    global_quantities = ()

    for key, value in lightcone_q.items():
        if value is True:
            lightcone_quantities += (key,)
    
    for key, value in global_q.items():
        if value is True:
            global_quantities += (key,)

    ## Set default values to output if nothing is set in the config file
    if len(lightcone_quantities) == 0:
        lightcone_quantities = ('xH_box',)

    if len(global_quantities) == 0:
        global_quantities = ('brightness_temp', 'xH_box',)

    try: 

        if PY21CMFAST:
            lightcone = p21f.run_lightcone(
                    user_params          = user_params,
                    astro_params         = astro_params,
                    cosmo_params         = cosmo_params,
                    flag_options         = flag_options,
                    lightcone_quantities = lightcone_quantities,
                    global_quantities    = global_quantities,
                    direc                = cache_path, 
                    random_seed          = random_seed,
                    **extra_params,
                    **kwargs,
                )
        else:
            ImportError("Need 21cmFAST to run ligthcones")
        
    except Exception as e :

        print(e)

        lightcone = None
        run_id    = None

    try: 
        # at the end, we clear the cache 
        if PY21CMFAST:
            p21f.cache_tools.clear_cache(direc=cache_path)
        else:
            ImportError("Need 21cmFAST to run ligthcones")
        # delete the directory once it has been emptied
        os.rmdir(cache_path) 
    except FileNotFoundError:
        pass

    return lightcone, run_id, output_dir


    
 


##################################################################################
# This file is part of 21cmCAST.
#
# Copyright (c) 2023, Gaétan Facchinetti
#
# 21cmCAST is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. 21cmCAST is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with 21cmCAST. 
# If not, see <https://www.gnu.org/licenses/>.
#
# -------------------------------------------------------------------------------
#
# Parts of this code have been copied and modified 
# from https://github.com/charlottenosam/21cmfish
# 
# - MIT License
# -
# - Copyright (c) 2019, Charlotte Mason
# - 
# - Permission is hereby granted, free of charge, to any person obtaining a copy
# - of this software and associated documentation files (the "Software"), to deal
# - in the Software without restriction, including without limitation the rights
# - to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# - copies of the Software, and to permit persons to whom the Software is
# - furnished to do so, subject to the following conditions:
# - 
# - The above copyright notice and this permission notice shall be included in all
# - copies or substantial portions of the Software.
# - 
# - THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# - IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# - FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# - AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# - LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# - OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# - SOFTWARE.
##################################################################################

import glob


import numpy as np
import pickle
from scipy   import interpolate
from astropy import units
import warnings

PY21CMFAST = True

try:
    import py21cmfast as p21f
except ImportError:
    PY21CMFAST = False
    p21f = None
    warnings.warn("21cmFAST could not be found! 21cmCAST can work on cached data but will not be able to process any new lightcones.")

from py21cmcast import power         as p21c_p
from py21cmcast import tools         as p21c_tools
from py21cmcast import experiments   as p21c_exp



def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    _filename = 'py21cmcast/' + filename.split("/")[-1]
    return ' %s:%s: %s (%s)\n' % (_filename, lineno, message, category.__name__)

warnings.formatwarning = warning_on_one_line


def compare_arrays(array_1 : np.ndarray, array_2 : np.ndarray, eps : float):
    """
    ## Compare arrays to a certain precision

    Parameters
    ----------
    array_1 : np.ndarray
        First array to compare
    array_1 : np.ndarray
        Second array to compare
    eps : float
        Relative precision asked for the comparision

    Returns:
    --------
    bool
        True if arrays are equal to the precision eps, False otherwise
    """

    if len(array_1) != len(array_2):
        return False
    
    return bool(np.all(2* np.abs((array_1 - array_2)/(array_1 + array_2)) < eps))
    



class Run:

    """
    # Class to define and analyse one run (and its associated lightcone)

    Fetch/Compute and store derived quantities such as global quantities
    or the power_spectrum of the brightness temperature.
    """

    def __init__(self, dir_path : str, lightcone_name : str,  z_bins : list = None, z_array : list = None, 
                k_bins : list = None, logk : bool = False, p : float = 0., *, lightcone = None, load : bool = True, save : bool = True, 
                 verbose : bool  = True) -> None: 
        """
        Parameters
        ----------
        dir_path : str
            Path to the directory where the lightcone is saved
        lightcone_name : str
            Name of the lightcone file to work with
        z_bins : list
            Array of redshift bin edges
        z_centers : list
            Array of redshift bin centers
        k_bins : list
            Array of mode bin edges
        logk : bool
            Indicates that the k mode edges are logarithmically distributed
        p : float, optional
            Associated value of the parameter associated to the run
        lightcone: lightcone object, optional
            21cmFAST lightcone object if not None, the lightcone is not 
            defined from `lightcone_name` but from this lightcone object
        load : bool, optional
            When True, first search for a cached precomputed tables of
            the quantities derived in the class (power_spectrum, noise, etc.)
            The cached tables have to be compatible with k_bins and z_bins
            (i.e. preomputed for the same arrays z_bins and k_bins)
            Default is True
        save : bool, optional
            When True, save the derived quantities in a cached table
            Default is True
        verbose : bool, optional
            When true, outputs information for the user
            Default is True
        """
        
        self._dir_path        = dir_path
        self._name            = lightcone_name
        self._filename_data   = self._dir_path + '/cache/Table_' + self._name + '.npz'    
        self._filename_params = self._dir_path + '/cache/Param_' + self._name + '.pkl' 

        # Create the database directory if it does not already exists
        p21c_tools.make_directory(self._dir_path + "/cache", clean_existing_dir=False)
        p21c_tools.make_directory(self._dir_path + "/power_spectra", clean_existing_dir=False)
        p21c_tools.make_directory(self._dir_path + "/derivatives", clean_existing_dir=False)
        p21c_tools.make_directory(self._dir_path + "/global_quantities", clean_existing_dir=False)

        self._verbose = verbose

        ## if load true we try loading precomputed tables
        if load is True : 

            _load_successfull = self._load()

            _params_match = True

            # If we load the file correctly and if the input parameter correspond
            # then we don't need to go further and can skip the full computation again
            if _load_successfull is True:
                if z_bins is not None:
                    if compare_arrays(self._z_bins, z_bins, 1e-5) is not True:
                        _params_match = False
                        warnings.warn("z-bins in input are different than the one used to precompute the tables")
                if k_bins is not None:
                    if compare_arrays(self._k_bins, k_bins, 1e-5) is not True:
                        _params_match = False
                        warnings.warn("k-bins in input are different than the one used to precompute the tables")
                if z_array is not None:
                    if compare_arrays(self._z_array, z_array, 1e-5) is not True:
                        _params_match = False
                        warnings.warn("z-array in input is different than the one used to precompute the tables")

                if _params_match is True:
                    return None
                else:
                    print("recomputing the tables with the new bins")

            #if (z_bins is None or k_bins is None or z_array is None) and _load_successfull is False:
            #    raise ValueError("Need to pass z_bins and k_bins as inputs")


        if load is False or _load_successfull is False or _params_match is False:

            self._z_bins    = z_bins
            self._k_bins    = k_bins
            self._z_array   = z_array
            self._logk      = logk
            self._p         = p

            # Get the power spectrum from the Lightcone
            if lightcone is not None:
                self._lightcone = lightcone
            else:  
                if PY21CMFAST:
                    self._lightcone   = p21f.LightCone.read(self._dir_path + "/" + self._name)
                else : 
                    ImportError("21cmFAST is needed to read new lightcones")

            self._astro_params = dict(self._lightcone.astro_params.self)
            self._user_params  = dict(self._lightcone.user_params.self)
            self._cosmo_params = dict(self._lightcone.cosmo_params.self)
            self._flag_options = dict(self._lightcone.flag_options.self)

            # Compute the power spectrum and fetch global quantities
            self._get_power_spectra()    
            self._get_global_quantities()

            # Compute the optical depth to reionization
            if PY21CMFAST:
                try: 
                    self._tau_ion = p21f.compute_tau(redshifts=self._z_glob, global_xHI = self.xH_box, user_params=self._user_params, cosmo_params=self._cosmo_params)
                except:
                    self._tau_ion = 0.0
                    if self._verbose:
                        warnings.warn("Impossible to compute the optical depth to reionization tau_ion")
                    
            else : 
                ImportError("21cmFAST is needed to evaluate the optical depth to reionization")

            if save is True:
                self._save()


    def _get_power_spectra(self):

        if ((self._z_bins is None) or (self._z_bins == np.array([]))
            or (self._k_bins is None) or (self._k_bins == np.array([]))
            or (self._z_array is None) or (self._z_array == np.array([]))):
            
            if self._verbose:
                warnings.warn("Impossible to compute the power spectra as bins are missing")
            
            self._power_spectrum = np.array([])
            self._ps_poisson_noise = np.array([])
            self._k_array = np.array([])
            self._chunk_indices = np.array([])

        try:

            self._lc_redshifts        = self._lightcone.lightcone_redshifts
            self._chunk_indices       = [np.argmin(np.abs(self._lc_redshifts - z)) for z in self._z_bins]
            _, _data_arr  = p21c_p.compute_powerspectra_1D(self._lightcone, chunk_indices = self._chunk_indices, 
                                                                    n_psbins=self._k_bins.value, logk=self._logk, 
                                                                    remove_nans=False, vb=False)
            
            self._power_spectrum   = np.array([data['delta'] for data in _data_arr])
            self._ps_poisson_noise = np.array([data['err_delta'] for data in _data_arr])

            _k_array = np.array([data['k'] for data in _data_arr])
            assert np.all(np.abs(np.diff(_k_array, axis=0)[0]/_k_array) <= 1e-3)
            self._k_array = _k_array[0]

        except Exception as e:
            if self._verbose:
                warnings.warn("Impossible to compute the power spectra: \n" + str(e))

            self._power_spectrum = np.array([])
            self._ps_poisson_noise = np.array([])
            self._k_array = np.array([])
            self._chunk_indices = np.array([])


    
    def _get_global_quantities(self):

        _lc_glob_redshifts = self._lightcone.node_redshifts
        self._z_glob       = np.linspace(_lc_glob_redshifts[-1], _lc_glob_redshifts[0], 500)
  
        _global_signal = self._lightcone.global_quantities.get('brightness_temp', np.zeros(len(_lc_glob_redshifts), dtype=np.float64))
        _xH_box        = self._lightcone.global_quantities.get('xH_box', np.zeros(len(_lc_glob_redshifts), dtype=np.float64))
        _x_e_box       = self._lightcone.global_quantities.get('x_e_box', np.zeros(len(_lc_glob_redshifts), dtype=np.float64))
        _Ts_box        = self._lightcone.global_quantities.get('Ts_box', np.zeros(len(_lc_glob_redshifts), dtype=np.float64))
        _Tk_box        = self._lightcone.global_quantities.get('Tk_box', np.zeros(len(_lc_glob_redshifts), dtype=np.float64))

        self._global_signal = interpolate.interp1d(_lc_glob_redshifts, _global_signal)(self._z_glob)
        self._xH_box        = interpolate.interp1d(_lc_glob_redshifts, _xH_box)(self._z_glob)
        self._x_e_box       = interpolate.interp1d(_lc_glob_redshifts, _x_e_box)(self._z_glob)
        self._Ts_box        = interpolate.interp1d(_lc_glob_redshifts, _Ts_box)(self._z_glob)
        self._Tk_box        = interpolate.interp1d(_lc_glob_redshifts, _Tk_box)(self._z_glob)


    def _save(self):
        """
        ##  Saves all the attributes of the class to be easily reload later if necessary

        numpy arrays are saved in an .npz format
        scalar parameters / attributes are saved in a dictionnary in a .pkl format
        """

        with open(self._filename_data, 'wb') as file: 
            np.savez(file, power_spectrum     = self.power_spectrum,
                            ps_poisson_noise  = self.ps_poisson_noise,
                            global_signal     = self.global_signal,
                            chunk_indices     = self.chunk_indices,
                            xH_box            = self.xH_box,
                            x_e_box           = self.x_e_box,
                            Ts_box            = self.Ts_box,
                            Tk_box            = self.Tk_box,
                            z_array           = self.z_array if self.z_array is not None else np.array([]),
                            z_bins            = self.z_bins  if self.z_bins is not None else np.array([]),
                            k_bins            = self.k_bins  if self.k_bins is not None else np.array([]),
                            k_array           = self.k_array,
                            z_glob            = self.z_glob, 
                            tau_ion           = self.tau_ion)
        
        # Prepare the dictionnary of parameters
        param_dict = {'logk' : self.logk, 'p' : self.p, 
                    'astro_params': self.astro_params, 
                    'user_params': self.user_params,
                    'flag_options': self.flag_options,
                    'cosmo_params' : self.cosmo_params}
        
        with open(self._filename_params, 'wb') as file:
            pickle.dump(param_dict, file)
        

    def _load(self):
        """
        ##  Loads all the attributes of the class
        """

        data   = None
        params = None

        try:

            with open(self._filename_data, 'rb') as file: 
                data = np.load(file)
                
                self._power_spectrum    = data['power_spectrum']
                self._ps_poisson_noise  = data['ps_poisson_noise']
                self._chunk_indices     = data['chunk_indices']
                self._z_array           = data['z_array']
                self._k_array           = data['k_array']
                self._z_bins            = data['z_bins']
                self._k_bins            = data['k_bins']  / units.Mpc
                self._global_signal     = data['global_signal']
                self._xH_box            = data['xH_box']
                self._x_e_box           = data.get('x_e_box', None)
                self._Ts_box            = data.get('Ts_box', None)
                self._Tk_box            = data.get('Tk_box', None)
                self._z_glob            = data['z_glob']
                self._tau_ion           = data['tau_ion'].item()


            with open(self._filename_params, 'rb') as file:
                params = pickle.load(file)

                self._logk          = params['logk']
                self._p             = params['p']
                self._astro_params  = params['astro_params']
                self._user_params   = params['user_params']
                self._flag_options  = params['flag_options']  
                self._cosmo_params  = params['cosmo_params']

            return True

        except FileNotFoundError:
            if self._verbose is True:
                print("No existing cached data found for " + self._name)
            
            return False
        

    def plot_power_spectrum(self, std = None, figname = None, plot=True, ps_modeling_noise=None) :  

        error = self.ps_poisson_noise
        if ps_modeling_noise is not None :
            error = np.sqrt(self.ps_poisson_noise**2 + self.ps_modeling_noise**2)

        fig = p21c_tools.plot_func_vs_z_and_k(self.z_array, self.k_array, self.power_spectrum, 
                                                std = error, title=r'$\Delta_{21}^2 ~{\rm [mK^2]}$', 
                                                xlim = [self._k_bins[0].value, self._k_bins[-1].value], xlog=self._logk, ylog=True)
        
        if plot is True : 

            if figname is None:
                figname = self._dir_path + "/power_spectra/power_spectrum.pdf"
        
            fig.savefig(figname, bbox_layout='tight')
            p21c_tools.close_figure(fig)

        return

    @property
    def power_spectrum(self):
        return self._power_spectrum

    @property
    def ps_poisson_noise(self):
        return self._ps_poisson_noise
    
    @property
    def global_signal(self):
        return self._global_signal
    
    @property
    def xH_box(self):
        return self._xH_box
    
    @property
    def x_e_box(self):
        return self._x_e_box
    
    @property
    def Ts_box(self):
        return self._Ts_box

    @property
    def Tk_box(self):
        return self._Tk_box
    
    @property
    def z_glob(self):
        return self._z_glob

    @property
    def z_bins(self):
        return self._z_bins

    @property
    def k_bins(self): 
        return self._k_bins

    @property
    def z_array(self):
        return self._z_array

    @property
    def k_array(self):
        return self._k_array

    @property
    def logk(self): 
        return self._logk

    @property
    def chunk_indices(self):
        return self._chunk_indices

    @property
    def p(self):
        return self._p

    ## Lighcone properties
    @property
    def astro_params(self):
        return self._astro_params

    @property
    def user_params(self):
        return self._user_params

    @property
    def flag_options(self):
        return self._flag_options

    @property
    def cosmo_params(self):
        return self._cosmo_params

    @property
    def tau_ion(self):
        return self._tau_ion





class Fiducial(Run): 
    """
    # Class to define the Fiducial run
    """

    def __init__(self, dir_path : str, z_bins : list, z_array : list, k_bins : list, logk : bool,
                frac_noise : float = 0.2, rs : int = None, ps_modeling_noise : np.ndarray = None, 
                load : bool = True, save : bool = True, verbose : bool = False, **kwargs) -> None:

        """
        Parameters
        ----------
        dir_path : str
            Path to the directory where the FIDUCIAL lightcone is saved
        z_bins : list
            Array of redshift bin edges
        z_array : list
            Array of redshift bin centers
        k_bins : list
            Array of mode bin edges
        logk : bool
            Indicates that the k mode edges are logarithmically distributed
        frac_noise : float, optional
            Percentage level of modeling noise to add to the power spectrum
            Default is 0.2
        rs : int, optional
            Value of the random seed used in 21cmFAST to compute the FIDUCIAL
            Default is None
        ps_modeling_noise : np.ndarray, optional
            Overrides frac_noise by giving the modelign noise in a 2D table depending
            on the redshift bin and the mode bin (of the same size as the power_spectrum).
            Default is None
        load : bool, optional
            When True, first search for a cached precomputed tables of
            the quantities derived in the class (power_spectrum, noise, etc.)
            The cached tables have to be compatible with k_bins and z_bins
            (i.e. preomputed for the same arrays z_bins and k_bins)
            Default is True
        save : bool, optional
            When True, save the derived quantities in a cached table
            Default is True
        verbose : bool, optional
            When true, outputs information for the user
            Default is True

   
        """

        self._dir_path             = dir_path
        self._filename_exp_noise   = self._dir_path + '/cache/Table_exp_noise'
        self._verbose              = verbose

        # get the lightcones filenames
        _loading_success = True
        if load is True:
            # here we look in the cache folder
            _str_file_name = self._dir_path + "/cache/Table_Lightcone_rs*_FIDUCIAL.h5.npz" if (rs is None) else self._dir_path + "/cache/Table_Lightcone_rs" + str(rs) + "_FIDUCIAL.h5.npz"
            _lightcone_file_name = glob.glob(_str_file_name)
            
            if len(_lightcone_file_name) == 1:
                _loading_success = True
                _file_name = (_lightcone_file_name[0].split('/')[-1])[6:-4] # remove the dir_path and 'Table_' ... '.npz'
            else:
                _loading_success = False

          # if we do not load, one has to find a lightcone for the fiducial
        if load is False or _loading_success is False : 
            _str_file_name = self._dir_path + "/Lightcone_rs*_FIDUCIAL.h5" if (rs is None) else self._dir_path + "/Lightcone_rs" + str(rs) + "_FIDUCIAL.h5"
            _lightcone_file_name = glob.glob(_str_file_name)
            _file_name = _lightcone_file_name[0].split('/')[-1] 

        assert len(_lightcone_file_name) == 1, 'No fiducial lightcone found or too many'
        
        self._rs   = _file_name.split('_')[-2][2:]

        # Initialise the parent object
        super().__init__(self._dir_path, _file_name, z_bins, z_array, k_bins, logk, 0., lightcone = None, load=load, save=save, verbose = verbose)
        
        # Getting the noise associated to the fiducial
        self._frac_noise         = frac_noise
        self._ps_modeling_noise  = ps_modeling_noise if (ps_modeling_noise is not None) else self._frac_noise * self._power_spectrum
        self._astro_params       = self._astro_params

        self._is_ps_sens_computed = False


    def _save_ps_exp_noise(self, observation):
         
         with open(self._filename_exp_noise + '_' + observation + '.npz', 'wb') as file: 
            np.savez(file, ps_exp_noise         = self.ps_exp_noise,
                            z_array             = self.z_array,
                            k_array             = self.k_array,
                            z_bins              = self.z_bins,
                            k_bins              = self.k_bins)
            

    def _load_ps_exp_noise(self, observation_set):

        data   = None

        try:

            with open(self._filename_exp_noise + '_' + observation_set + '.npz' , 'rb') as file: 
                data = np.load(file)

                _z_array           = data['z_array']
                _k_array           = data['k_array']
                _z_bins            = data['z_bins']
                _k_bins            = data['k_bins']  / units.Mpc

                # if the binning are compatibles then we can load the sensitivity
                if compare_arrays(_z_array, self._z_array, 1e-3) \
                    and compare_arrays(_k_array, self._k_array, 1e-3) \
                    and compare_arrays(_z_bins, self._z_bins, 1e-3) \
                    and compare_arrays(_k_bins, self._k_bins, 1e-3) :

                    self._ps_exp_noise      = data['ps_exp_noise']
                else:
                    return False

            return True

        except FileNotFoundError:
            if self._verbose is True:
                warnings.warn("No existing data found for " + observation_set + 'with this binning')

            return False

    @property
    def dir_path(self):
        return self._dir_path

    @property
    def astro_params(self):
        return self._astro_params

    @property
    def frac_noise(self):
        return self._frac_noise

    @frac_noise.setter
    def frac_noise(self, value):
        if value != self._frac_noise:
            warnings.warn("frac noise and modeling noise have been changed, all related quantities should be recomputed")
            self._frac_noise        = value
            self._ps_modeling_noise = self._frac_noise * self._power_spectrum

    @property
    def ps_exp_noise(self):
        return np.array(self._ps_exp_noise)
    
    @property
    def ps_modeling_noise(self):
        return self._ps_modeling_noise

    @ps_modeling_noise.setter
    def ps_modeling_noise(self, value):
        if value != self._ps_modeling_noise:
            warnings.warn("Modeling noise has been changed, all related quantities should be recomputed")
            warnings.warn("Changind modeling noise directly is not advised, you should maybe change it through frac_noise")
            self._ps_modeling_noise = value
    
    @property
    def rs(self):
        return self._rs
    
    @property
    def k_array_sens(self):
        return self._k_array_sens
    
    @property
    def power_spectrum_sens(self):
        return self._power_spectrum_sens
    

    def compute_power_spectrum_sensitivity(self):

        # Get the Lightcone quantity
        if PY21CMFAST:
            self._lightcone   = p21f.LightCone.read(self._dir_path + "/" + self._name)
        else : 
            ImportError("21cmFAST is needed to read new lightcones")

        # In order to compute the sensitivity we need the full range of z and k possible
        # We recompute a second power spectra for the sensitivity on a broader range
        # We use the same redshift chunks but let the array of mode free (and here logarithmically spaced by default)
        # This makes the code longer but it is necessary to have everything well defined and no numerical problems
        _, _data_arr   = p21c_p.compute_powerspectra_1D(self._lightcone, chunk_indices = self._chunk_indices, remove_nans=False, vb=False)

        self._k_array_sens         = np.array([data['k'] for data in _data_arr])
        self._power_spectrum_sens  = np.array([data['delta'] for data in _data_arr])

        self._is_ps_sens_computed = True



    def compute_sensitivity(self, observation_set = None, load = True, save = True):

        # Getting the name of the observation settings
        obs_name = "default_HERA" if (observation_set is None) else observation_set.name

        # Trying to load the senstivity from the observations sent
        # Here we just need the name of the observation object
        _load_succesfull = self._load_ps_exp_noise(obs_name) if load else False

        if _load_succesfull:
            print("Successfully loaded sensitivity for ", obs_name)
            return

        # If we cannot load then we need 21cmSense, make sure that we can find the module
        if p21c_exp.PY21CMSENSE is False:
            raise ImportError("The module 21cmSense could not be imported or was not found, sensitivity cannot be computed")
    
        # if observation_set is None we initialise it to default HERA
        _observation_set = p21c_exp.default_observation_set(self.z_array) if (observation_set is None) else observation_set
        
        assert (_observation_set.observations is not None), ValueError("observations in ObservationSet must be a set of observations from 21cmSense")
        assert (len(_observation_set.observations) == len(self.z_array)), ValueError("We must have one observation for each redshift bin") 

        _std = None
        self._ps_exp_noise = None

        # if the power spectrum is not computed we do compute it
        if self._is_ps_sens_computed is False:
            self.compute_power_spectrum_sensitivity()

        _std = [p21c_exp.extract_noise_from_fiducial(self.k_array_sens[iz], self.power_spectrum_sens[iz], self.k_array, _observation_set.observations[iz]) for iz, _ in enumerate(self.z_array)]

        self._ps_exp_noise = _std
            
        if save is True:
            self._save_ps_exp_noise(_observation_set.name)


    
    def chi2_UV_luminosity_functions(self, data_set = 'Bouwens21', plot = True):

        z_uv, m_uv, l_uv, sigma_l_uv   = p21c_tools.load_uv_luminosity_functions(data_set)
        
        if PY21CMFAST:
            m_uv_sim , _ , log10_l_uv_sim  = p21f.compute_luminosity_function(redshifts = z_uv, 
                                                    user_params  = self._user_params, 
                                                    astro_params = self._astro_params, 
                                                    flag_options = self._flag_options)
        else : 
            ImportError("21cmFAST is needed to compute the UV luminosity functions")

        
        if plot is True:
            fig = p21c_tools.plot_func(m_uv_sim, 10**log10_l_uv_sim, ylog=True, xlim=[-14, -25], ylim =[1e-12, 1], xlabel=r'$M_{\rm UV}$', ylabel=r'$\phi_{\rm UV}$')
            for iz, z in enumerate(z_uv) :
                    fig.gca().errorbar(m_uv[iz], l_uv[iz], yerr=sigma_l_uv[iz], linestyle='', capsize=2, elinewidth=0.5, label=r'${}$'.format(z))

            fig.gca().legend()
            fig.savefig(self._dir_path + '/global_quantities/UV_lum_func_FIDUCIAL.pdf', bbox_inches='tight')
            p21c_tools.close_figure(fig)

        ## The chi2 is given by a sum on the elements where z > 6 (as we cannot trust 21cmFAST below)

        _chi2 = 0
        _nval = 0
        for iz, z in enumerate(z_uv):
            if z > 6:
                log10_l_uv_func = interpolate.interp1d(m_uv_sim[iz], log10_l_uv_sim[iz])
                for im, m in enumerate(m_uv[iz]) :     
                    _chi2 = _chi2 + (10**log10_l_uv_func(m) - l_uv[iz][im])**2/(sigma_l_uv[iz][im]**2)
                    _nval = _nval + 1

        ## Now compute the reduced chi2
        _reduced_chi2 = _chi2 / _nval

        return _reduced_chi2
    

    def plot_power_spectrum(self):
        super().plot_power_spectrum(std=self._ps_exp_noise, figname = self._dir_path + "/power_spectra/power_spectrum_FIDUCIAL.pdf")


    def plot_xH_box(self):
        fig = p21c_tools.plot_func(self.z_glob, self.xH_box,
                                    xlabel=r'$z$',
                                    ylabel=r'$x_{\rm H_{I}}$')
        fig.savefig(self._dir_path + '/global_quantities/xH_FIDUCIAL.pdf', bbox_inches='tight')
        p21c_tools.close_figure(fig)


    def plot_global_signal(self):
        fig = p21c_tools.plot_func(self.z_glob, self.global_signal, ylim=[-150, 50],
                                    xlabel=r'$z$',
                                    ylabel=r'$\overline{T_{\rm b}}~\rm [mK]$')
        fig.savefig(self._dir_path + '/global_quantities/global_signal_FIDUCIAL.pdf', bbox_inches='tight')
        p21c_tools.close_figure(fig)





class Parameter:

    def __init__(self, fiducial, name, plot = True, verbose = True, values = None, **kwargs):
        
        self._fiducial       = fiducial
        self._name           = name
        self._plot           = plot

        self._dir_path       = self._fiducial.dir_path
        self._astro_params   = self._fiducial.astro_params
        self._cosmo_params   = self._fiducial.cosmo_params
        self._z_bins         = self._fiducial.z_bins
        self._z_array        = self._fiducial.z_array
        self._k_bins         = self._fiducial.k_bins
        self._logk           = self._fiducial.logk

        self._verbose        = verbose

        # Additional parameters to specify the value of the parameter
        self._values         =  values
        self._add_name       =  kwargs.get('add_name', '')
        
        if self._add_name != '':
            self._add_name = '_' + self._add_name
            
        __params_plot  = p21c_tools._PARAMS_PLOT.get(self._name, None)
        self._tex_name = __params_plot['tex_name'] if (__params_plot is not None) else p21c_tools._PARAMS_PLOT.get('theta', None)['tex_name']
       
        self._load = kwargs.get('load', True)

        if (name not in self._astro_params) or (name not in self._cosmo_params):
            ValueError("ERROR: the name does not corresponds to any parameter")

        if name in self._astro_params:
            self._param_fid = self._astro_params[self._name]
        
        if name in self._cosmo_params:
            self._param_fid = self._cosmo_params[self._name]

        _loading_success = True

        ############################################################
        # get the lightcones / saved cached values from the filenames
        if self._load is True:
              
            _file_name = self._dir_path + "/cache/Table_Lightcone_rs" + str(self._fiducial.rs) + "_*" + self._name + self._add_name + "_"
            
            
            if self._values is None :
                # if we do not specify values for the parameter
                _lightcone_file_name = glob.glob(_file_name + "[0-9\-]*" + ".h5.npz")
                _lightcone_file_name = [fname[:-4] for fname in _lightcone_file_name] 
            else: 
                # if we specify a value for the parameter
                if not isinstance(self._values, list):
                    self._values = [self._values]
                _lval = [None] * len(self._values)
                
                for ival, val in enumerate(self._values):
                    _lval[ival] = glob.glob(_file_name + '{:.4e}'.format(val) + ".h5.npz")

                if len(_lval[0]) > 0:
                    _lightcone_file_name = [lval[0][:-4] for lval in _lval] 
                else:
                    _lightcone_file_name = []
            
            # test if saved lightcones have been found successfully
            _loading_success = True if len(_lightcone_file_name) > 0 else False

        if self._load is False or _loading_success is False:
            
            _file_name = self._dir_path + "/Lightcone_rs" + str(self._fiducial.rs) + "_*" + self._name + self._add_name + "_"
            if self._values is None :
                _lightcone_file_name = glob.glob(_file_name + "[0-9\-]*" + ".h5")
            else: 
                
                if not isinstance(self._values, list):
                    self._values = [self._values]
                _lval = [None] * len(self._values)
                
                for ival, val in enumerate(self._values):
                    _lval[ival] = glob.glob(_file_name + '{:.4e}'.format(val) + ".h5")
                    assert len(_lval[0]) > 0,  FileNotFoundError(f'No files found for the lightcone associated to ' + self._name + self._add_name + f' variations for value: ' + '{:.4e}'.format(val))
                
                _lightcone_file_name = [lval[0] for lval in _lval]        
             
        assert len(_lightcone_file_name) > 0,  FileNotFoundError(f'No files found for the lightcone associated to ' + str(self._name) + f' variations')
        ############################################################

        # get (from the filenames) the quantity by which the parameter has been varies from the fiducial
        self._p_value = []
        for file_name in _lightcone_file_name:
            # Get the value of p from the thing
            self._p_value.append(float(file_name.split("_")[-1][:-3]))

        # If more than one file with the same p_values we remove all identical numbers
        self._p_value = list(set(self._p_value))

        # Sort the p_values from the smallest to the largest
        self._p_value = np.sort(self._p_value)

        # Check that the p_values are consistant
        assert len(self._p_value) == 1 or (len(self._p_value) == 2 and ((self._p_value[0] - self._param_fid) * (self._p_value[1] - self._param_fid)) < 0), "p_value : " + str(self._p_value)

        if verbose is True: 
            print("------------------------------------------")
            print(self._name  + " has been varied with p = " + str(self._p_value))
            print("Loading the lightcones and computing the power spectra")
        else :
            if self._add_name != '':
                print("Treating parameter " + self._name + ' (' + self._add_name[1:] + ')')
            else:
                print("Treating parameter " + self._name)

        # We get the lightcones and then create the corresponding runs objects
        self._runs =  [Run(self._dir_path, 
                                    'Lightcone_rs' + str(self._fiducial.rs) + '_' + self._name + self._add_name + '_{:.4e}'.format(p) + '.h5', 
                                    self._z_bins, self._z_array, self._k_bins, 
                                    self._logk, p, **kwargs) for p in self._p_value]

        if verbose is True: 
            print("Power spectra of " + self._name  + " computed")
      
        ## Check that the k-arrays, z-arrays, k-bins and z-bins correspond 
        for run in self._runs:
            assert compare_arrays(run.z_array, self._fiducial.z_array, 1e-2)
            assert compare_arrays(run.k_array, self._fiducial.k_array, 1e-2)
            assert compare_arrays(run.z_bins,  self._fiducial.z_bins,  1e-2)
            assert compare_arrays(run.k_bins,  self._fiducial.k_bins,  1e-2)

        ## Define unique k and z arrays
        self._k_array = self._fiducial.k_array
        self._z_array = self._fiducial.z_array
        self._k_bins  = self._fiducial.k_bins
        self._z_bins  = self._fiducial.z_bins

        if verbose is True: 
            print("Computing the derivatives of " + self._name)

        self.compute_ps_derivative()

        if verbose is True: 
            print("Derivative of " + self._name  + " computed")
            print("------------------------------------------")
            
        # Plotting the derivatives
        if self._plot is True:

            fig_der = self.plot_ps_derivative()
            fig_w_der = self.plot_weighted_ps_derivative()
            fig_ps = self.plot_power_spectra()

            fig_der.savefig(self._dir_path + "/derivatives/derivatives_" + self._name + self._add_name + ".pdf")
            fig_w_der.savefig(self._dir_path + "/derivatives/weighted_derivatives_" + self._name + self._add_name + ".pdf")
            fig_ps.savefig(self._dir_path + "/power_spectra/power_spectra_" + self._name + self._add_name + ".pdf")

            p21c_tools.close_figure(fig_der)
            p21c_tools.close_figure(fig_w_der)
            p21c_tools.close_figure(fig_ps)


    @property
    def ps_derivative(self):
        return self._ps_derivative

    @property
    def k_array(self):
        return self._k_array

    @property
    def z_array(self):
        return self._z_array

    @property
    def k_bins(self):
        return self._k_bins

    @property
    def z_bins(self):
        return self._z_bins

    @property
    def fiducial(self):
        return self._fiducial

    @property
    def name(self):
        return self._name


    def compute_ps_derivative(self):
        
        _der = [None] * len(self._z_array)
        
        #_param_fid = self._astro_params[self._name]

        # get all the parameters and sort them
        _params = np.zeros(len(self._runs))

        for irun, run in enumerate(self._runs):
            _params[irun] = run.p

        _params         = np.append(_params, self._param_fid)
        _params_sorted  = np.sort(_params)
        _mixing_params  = np.argsort(_params)


        # loop over all the redshift bins
        for iz, z in enumerate(self._z_array) :   

            # get an array of power spectra in the same order as the parameters
            _power_spectra = [run.power_spectrum[iz] for run in self._runs]
            _power_spectra.append(self._fiducial.power_spectrum[iz])

            # rearrange the power spectra in the same order of the parameters
            _power_spectra_sorted = np.array(_power_spectra)[_mixing_params]        

            # evaluate the derivative as a gradient
            _der[iz] = np.gradient(_power_spectra_sorted, _params_sorted, axis=0)

        # arrange the derivative whether they are left, right or centred
        self._ps_derivative  = {'left' : None, 'right' : None, 'centred' : None}

        if len(self._p_value) == 2:
            self._ps_derivative['left']    = [_der[iz][0] for iz, _ in enumerate(self._z_array)]
            self._ps_derivative['centred'] = [_der[iz][1] for iz, _ in enumerate(self._z_array)]
            self._ps_derivative['right']   = [_der[iz][2] for iz, _ in enumerate(self._z_array)]

        if len(self._p_value) == 1 and self._p_value[0] < 0 :
            self._ps_derivative['left'] = [_der[iz][0] for iz, _ in enumerate(self._z_array)]
        
        if len(self._p_value) == 1 and self._p_value[0] > 0 :
            self._ps_derivative['right'] = [_der[iz][0] for iz, _ in enumerate(self._z_array)]

        

    def weighted_ps_derivative(self, kind: str ='centred'):
       
        """
        ## Weighted derivative of the power spectrum with respect to the parameter
        
        Params:
        -------
        kind : str, optional
            choice of derivative ('left', 'right', or 'centred')

        Returns:
        --------
        The value of the derivative devided by the error
        """
        
        # experimental error
        ps_exp_noise   = self._fiducial.ps_exp_noise  

        # theoretical uncertainty from the simulation              
        ps_poisson_noise  = self._fiducial.ps_poisson_noise 
        ps_modeling_noise = self._fiducial.ps_modeling_noise
    
        
        # if fiducial as no standard deviation defined yet, return None
        if ps_exp_noise is None:
            return None

        # Initialize the derivative array to None
        der = None

        # If two sided derivative is asked then we are good here
        if kind == 'centred' :
            der = self._ps_derivative.get('centred', None)
        
        # Value to check if we use the one sided derivative 
        # by choice or because we cannot use the two-sided one
        _force_to_one_side = False

        # If there is no centred derivative
        if der is None or kind == 'left':
            if der is None and kind != 'left' and kind != 'right': 
                # Means that we could not read a value yet 
                # but not that we chose to use the left
                _force_to_one_side = True
            der = self._ps_derivative.get('left', None)
        
        if der is None or kind == 'right':
            if der is None and kind != 'right' and kind != 'left': 
                # Means that we could not read a value yet 
                # but not that we chose to use the right
                _force_to_one_side = True
            der = self._ps_derivative.get('right', None)

        if _force_to_one_side is True and self._verbose is True:
            print("Weighted derivative computed from the one_sided derivative")

        # We sum (quadratically) the two errors
        return der / np.sqrt(ps_exp_noise**2 + ps_poisson_noise**2 + ps_modeling_noise**2)  


    def plot_ps_derivative(self):

        der_array = []
        for key in self._ps_derivative.keys():
            if self._ps_derivative[key] is not None:
                der_array.append( self._ps_derivative[key])
        
        fig = p21c_tools.plot_func_vs_z_and_k(self._z_array, self._k_array, der_array, marker='.', markersize=2, 
                                                title=r'$\frac{\partial \Delta_{21}^2}{\partial ' + self._tex_name + r'}$', 
                                                xlim = [0.1, 1], xlog=self._logk, ylog=False)
        return fig


    def plot_power_spectra(self, **kwargs):

        _ps        = [self._fiducial.power_spectrum]
        _p_vals    = [self._param_fid]

        for run in self._runs:
            _ps.append(run.power_spectrum)
            _p_vals.append(run.p)

        _order     = np.argsort(_p_vals)
        _ps        = np.array(_ps)[_order]

        error_std = np.sqrt(self._fiducial.ps_exp_noise**2 + self._fiducial.ps_poisson_noise**2 + self.fiducial.ps_modeling_noise**2)

        fig = p21c_tools.plot_func_vs_z_and_k(self.z_array, self.k_array, _ps, std = error_std,
                                                title=r'$\Delta_{21}^2 ~ {\rm [mK^2]}$', 
                                                xlog=self._logk, ylog=True, istd = _order[0], **kwargs)

        return fig


    def plot_weighted_ps_derivative(self):

        if self.fiducial.ps_exp_noise is None:
            ValueError("Error: cannot plot the weighted derivatives if the error \
                        if the experimental error is not defined in the fiducial")


        der_array = [self.weighted_ps_derivative(kind=key) for key in self._ps_derivative.keys()]
        fig = p21c_tools.plot_func_vs_z_and_k(self._z_array, self._k_array, der_array, marker='.', markersize=2, 
                                                title=r'$\frac{1}{\sigma}\frac{\partial \Delta_{21}^2}{\partial ' + self._tex_name + r'}$', 
                                                xlim = [0.1, 1], xlog=self._logk, ylog=False)
     
        return fig




def evaluate_fisher_matrix(parameters):
    """
    ## Fisher matrix evaluator

    Params:
    -------
    parameters: array of Parameters objects
        parameters on which to compute the Fisher matrix
    
    Returns:
    --------
    dictionnary with keys: 'matrix' for the matrix itself and 'name' for 
    the parameter names associated to the matrix in the same order
    """
    
    # Get the standard deviation
    n_params = len(parameters)
    fisher_matrix = np.zeros((n_params, n_params))

    name_arr     = [''] * n_params
    weighted_der = [None] * n_params

    for ip, param in enumerate(parameters):
        name_arr[ip]      = param.name
        weighted_der[ip]  = param.weighted_ps_derivative()

    for i in range(0, n_params) :
        for j in range(0, n_params) :        
            fisher_matrix[i][j] = np.nansum(weighted_der[i] * weighted_der[j])
            
    return {'matrix' : fisher_matrix, 'name' : name_arr}
    

    




class CombinedRuns:
    """
    ## Smart collection of the same runs with different random seeds
    """

    def __init__(self, dir_path, name, z_bins = None, k_bins = None, logk=False, p : float = 0, save=True, load=True, verbose = True, **kwargs) -> None:
        
        self._z_bins  = z_bins
        self._k_bins  = k_bins
        self._logk    = logk
        self._p       = p

        # fetch all the lightcone files that correspond to runs the same parameters but different seed
        _lightcone_file_name = glob.glob(self._dir_path + "/Lightcone_*" + self._name + ".h5")
        
        if len(_lightcone_file_name) > 1:
            print("For " + self._name + ": grouping a total of " + str(len(_lightcone_file_name)) + " runs")
        
        # Create the array of runs
        self._runs =  [Run(dir_path, file_name, self._z_bins, self._k_bins, logk, p) for file_name in _lightcone_file_name]

        assert len(self._runs) > 0, "ERROR when searching for lightcones with a given name" 

        ## check that there is no problem and all z- and k- arrays have the same properties
        for irun in range(1, len(self._runs)) :

            # check that all with the same q-value have the same bins 
            assert compare_arrays(self._runs[0].k_array, self._runs[irun].k_array, 1e-5) 
            assert compare_arrays(self._runs[0].z_array, self._runs[irun].z_array, 1e-5) 
            assert compare_arrays(self._runs[0].k_bins,  self._runs[irun].k_bins,  1e-5) 
            assert compare_arrays(self._runs[0].z_bins,  self._runs[irun].z_bins,  1e-5) 
            
            # check that all with the same q-value have the same astro_params
            assert self._runs[0].astro_params  == self._runs[irun].astro_params
            assert self._runs[0].user_params   == self._runs[irun].user_params
            assert self._runs[0].flag_options  == self._runs[irun].flag_options
            assert self._runs[0].cosmo_params  == self._runs[irun].cosmo_params

        self._z_array = self._runs[0].z_array
        self._k_array = self._runs[0].k_array
        self._z_glob  = self._runs[0].z_glob

        self._average_quantities()


    def _average_quantities(self):
        
        ## compute the average values and the spread 
        self._power_spectrum    = np.average([run.power_spectrum for run in self._runs], axis=0)
        self._ps_poisson_noise  = np.average([run.ps_poisson_noise for run in self._runs], axis=0)
        self._ps_modeling_noise = np.std([run.power_spectrum for run in self._runs], axis=0)
        self._global_signal     = np.average([run.global_signal for run in self._runs], axis=0)
        self._xH_box            = np.average([run.xH_box for run in self._runs], axis=0)
        self._astro_params      = self._runs[0].astro_params 
        self._user_params       = self._runs[0].user_params
        self._flag_options      = self._runs[0].flag_options
        self._cosmo_params      = self._runs[0].cosmo_params

    @property
    def power_spectrum(self):
        return self._power_spectrum
    
    @property
    def ps_poisson_noise(self):
        return self._ps_poisson_noise

    @property
    def ps_modeling_noise(self):
        return self._ps_modeling_noise
    
    @property
    def global_signal(self):
        return self._global_signal
    
    @property
    def xH_box(self):
        return self._xH_box
    
    @property
    def z_glob(self):
        return self._z_glob

    @property
    def z_array(self):
        return self._z_array

    @property
    def k_array(self): 
        return self._k_array
    
    @property
    def z_bins(self):
        return self._z_bins

    @property
    def k_bins(self): 
        return self._k_bins
 
    @property
    def logk(self): 
        return self._logk

    @property
    def p(self):
        return self._p

    # lightcone properties
    @property
    def astro_params(self): 
        return self._astro_params
    
    @property
    def user_params(self):
        return self._user_params

    @property
    def flag_options(self):
        return self._flag_options

    @property
    def cosmo_params(self):
        return self._cosmo_params

    


from .experiments import (
    ObservationSet,
    default_HERA_observatories,
    default_observation_set,
    extract_noise_from_fiducial,
)

from .runs import (
    init_runs,
    init_grid_runs,
    read_grid_database,
    find_in_grid_database,
    make_config_one_varying_param,
    run_lightcone_from_config,
)

from .tools import (
    read_config_params,
    write_config_params,
    read_power_spectra,
    make_triangle_plot,
    plot_func_vs_z_and_k,
    plot_func,
    display_matrix,
    load_uv_luminosity_functions,
    prepare_plot,
    prepare_2subplots,
    prepare_triangle_plot,
)

from .power import (
    define_grid_modes_redshifts,
    generate_k_bins,
)

from .core import (
    evaluate_fisher_matrix,
    Fiducial,
    Run,
    CombinedRuns,
    Parameter,
)


##################################################################################
# This file is part of 21cmCAST.
#
# Copyright (c) 2023, Gaétan Facchinetti
#
# 21cmCAST is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. 21cmCAST is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with 21cmCAST. 
# If not, see <https://www.gnu.org/licenses/>.
#
# -------------------------------------------------------------------------------
#
# Parts of this code have been copied and modified 
# from https://github.com/charlottenosam/21cmfish
# 
# - MIT License
# -
# - Copyright (c) 2019, Charlotte Mason
# - 
# - Permission is hereby granted, free of charge, to any person obtaining a copy
# - of this software and associated documentation files (the "Software"), to deal
# - in the Software without restriction, including without limitation the rights
# - to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# - copies of the Software, and to permit persons to whom the Software is
# - furnished to do so, subject to the following conditions:
# - 
# - The above copyright notice and this permission notice shall be included in all
# - copies or substantial portions of the Software.
# - 
# - THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# - IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# - FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# - AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# - LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# - OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# - SOFTWARE.
##################################################################################



from matplotlib.gridspec import GridSpec
from matplotlib.lines import Line2D
from matplotlib.patches import Ellipse
import matplotlib.cm as mplt_cm
import matplotlib.pyplot as plt
import matplotlib.transforms as transforms

import numpy as np
import os
import shutil

import copy



def make_directory(path: str, clean_existing_dir:bool = False):
    
    if not os.path.exists(path): 
        os.mkdir(path)
    else:
        if clean_existing_dir is True:
            clean_directory(path)
        else:
            return True

    return False


def clean_directory(path: str):
    """ Clean the directory at the path: path """

    for filename in os.listdir(path):
        file_path = os.path.join(path, filename)
        
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)

            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))


def read_config_params(config_items, int_type = True, allow_lists = False):
    """
    Read ints and booleans from config files
    Use for user_params and flag_options only
    
    Parameters
    ----------
    item : str
        config dictionary item as a string
    Return
    ------
    config dictionary item as an int, bool or str
    """

    output_dict = dict()

    for key, value in dict(config_items).items():

        try:
            if allow_lists is True:
                cast_val = make_list(value, int_type)
            if allow_lists is False:
                if int_type is True:
                    cast_val = int(value)
                else:
                    cast_val = float(value)
        except:
            if value == 'True':
                cast_val =  True
            elif value == 'False':
                cast_val =  False
            else:   
                cast_val = value
    
        output_dict[key] = cast_val
        
    return output_dict


def make_list(str_input:str, int_type:bool = True):
    try:
        str_array = str_input.strip("[").strip("]").split(', ')
        if int_type is True:
            number_array = np.array([int(s) for s in str_array])
        else:
            number_array = np.array([float(s) for s in str_array])
        return number_array
    except:
        return str_input


def write_config_params(filename, name, output_dir, cache_dir, 
                        lightcone_quantities, global_quantities, 
                        extra_params, user_params, flag_options, 
                        astro_params, cosmo_params, key):

    with open(filename, 'w') as f:
       
        print("# Parameter file for : " + key, file = f)
        print('', file=f)

        print("[run]", file=f)
        print("name      : " + name, file=f)
        print("run_id    : " + key, file=f)
        print("output_dir : " + output_dir, file=f)
        print("cache_dir : " + cache_dir, file=f)
        print('', file=f)
        
        print("[extra_params]", file=f)
        
        for key, value in extra_params.items():
            print(key + " : " + str(value), file=f)

        
        print('', file=f)
        print("[lightcone_quantities]", file=f)

        for key, value in lightcone_quantities.items():
            print(key + " : " + str(value), file=f)

        print('', file = f)
        print("[global_quantities]", file=f)

        for key, value in global_quantities.items():
            print(key + " : " + str(value), file=f)

        print('', file=f)
        print("[user_params]", file=f)


        for key, value in user_params.items():
            print(key + " : " + str(value), file=f)

        print('', file=f)
        print("[flag_options]", file=f)

        for key, value in flag_options.items():
            print(key + " : " + str(value), file=f)

        print('', file=f)
        print("[astro_params]", file=f)

        for key, value in astro_params.items():
            print(key + " : " + str(value), file=f)

        print('', file=f)
        print("[cosmo_params]", file=f)

        for key, value in cosmo_params.items():
            print(key + " : " + str(value), file=f)



def read_power_spectra(folder_name: str):
    """ 
    Read the power spectra from a folder 
    The folder must contain a redshift array file: folder_name/redshift_chucnks.txt
    The power spectra must be labelled and organised as folder_name/ps_z_<<"{0:.1f}".format(z)>>.txt
    The units must be Mpc for k_arr, mK**2 for delta_arr and err_arr

    Parameters
    ----------
        folder_name: str
            path to the folder where the power_spectra_are_stored
    
    Returns
    -------
        z_arr: list[float]
            list of redshifts where the power_spectra are evaluated
        k_arr: list[list[float]] (Mpc^{-1})
            list of k values for every redshift
        delta_arr: list[list[float]] (mK^2)
            list of power_spectrum value for every redshift (in correspondance to k_arr)
        err_arr: list[list[float]]  (mK^2)
            list of the error on the power spectrum (in correspondance to k_arr and delta_arr)
    """

    z_arr     = np.genfromtxt(folder_name + '/power_spectra_vs_k/redshift_chunks.txt')

    k_arr     = []
    delta_arr = []
    err_arr   = []

    for iz, z in enumerate(z_arr):
        data = np.genfromtxt(folder_name + '/power_spectra_vs_k/ps_z_' + "{0:.1f}".format(z) + '.txt')

        k_arr.append(data[:, 0])
        delta_arr.append(data[:, 1])
        err_arr.append(data[:, 2])

    return z_arr, k_arr, delta_arr, err_arr






def confidence_ellipse(cov, mean_x, mean_y, ax, n_std=3.0, facecolor='none', **kwargs):
    """
    Create a plot of the covariance confidence ellipse of *x* and *y*.

    Parameters
    ----------
    x, y : array-like, shape (n, )
        Input data.

    ax : matplotlib.axes.Axes
        The axes object to draw the ellipse into.

    n_std : float
        The number of standard deviations to determine the ellipse's radiuses.

    **kwargs
        Forwarded to `~matplotlib.patches.Ellipse`

    Returns
    -------
    matplotlib.patches.Ellipse
    """

    pearson = cov[0, 1]/np.sqrt(cov[0, 0] * cov[1, 1])
    # Using a special case to obtain the eigenvalues of this
    # two-dimensionl dataset.
    ell_radius_x = np.sqrt(1 + pearson)
    ell_radius_y = np.sqrt(1 - pearson)
    ellipse = Ellipse((0, 0), width=ell_radius_x * 2, height=ell_radius_y * 2, facecolor=facecolor, **kwargs)

    # Calculating the stdandard deviation of x from
    # the squareroot of the variance and multiplying
    # with the given number of standard deviations.
    scale_x = np.sqrt(cov[0, 0]) * n_std

    # calculating the stdandard deviation of y ...
    scale_y = np.sqrt(cov[1, 1]) * n_std

    transf = transforms.Affine2D() \
        .rotate_deg(45) \
        .scale(scale_x, scale_y) \
        .translate(mean_x, mean_y)

    ellipse.set_transform(transf + ax.transData)
    return ax.add_patch(ellipse)



def ellipse_from_covariance(cov_matrix, fiducial):
    """ Returns arrays for drawing the covariance matrix
        This function is mainly used as a cross-check of confidence_ellipse()"""
    
    eigenvalues, eigenvectors = np.linalg.eig(cov_matrix)
    theta = np.linspace(0, 2*np.pi, 200)
    ellipse_x = (np.sqrt(eigenvalues[0])*np.cos(theta)*eigenvectors[0,0] + np.sqrt(eigenvalues[1])*np.sin(theta)*eigenvectors[0,1]) + fiducial[0]
    ellipse_y = (np.sqrt(eigenvalues[0])*np.cos(theta)*eigenvectors[1,0] + np.sqrt(eigenvalues[1])*np.sin(theta)*eigenvectors[1,1]) + fiducial[1]

    return ellipse_x, ellipse_y



_PARAMS_PLOT = {
    'theta'             : {'tex_name' : r'\theta'}, # default parameters
    'F_STAR10'          : {'tex_name' : r'\log_{10}[f_{\star, 10}]'},
    'F_STAR7_MINI'      : {'tex_name' : r'\log_{10}[f_{\star, 7}^{\rm mini}]'},
    'ALPHA_STAR'        : {'tex_name' : r'\alpha_{\star}', 'min': -0.7, 'max' : 1.7, 'ticks' : []},
    'ALPHA_STAR_MINI'   : {'tex_name' : r'\alpha_{\star}^{\rm mini}', 'min': -0.7, 'max' : 1.7, 'ticks' : []},
    't_STAR'            : {'tex_name' : r't_{\star}', 'min': -0.3, 'max' : 1.3, 'ticks' : []},
    'F_ESC10'           : {'tex_name' : r'\log_{10}[f_{\rm esc, 10}]', 'min': -2, 'max' : 0.1},
    'F_ESC7_MINI'       : {'tex_name' : r'\log_{10}[f_{\rm esc, 7}^{\rm mini}]', 'min': -2, 'max' : 0.1},
    'ALPHA_ESC'         : {'tex_name' : r'\alpha_{\rm esc}', 'min': -1.4, 'max': 0.6, 'ticks' : []},
    'L_X'               : {'tex_name' : r'\log_{10}\left[\frac{L_X}{\rm units}\right]', 'min': 40, 'max': 41},
    'L_X_MINI'          : {'tex_name' : r'\log_{10}\left[\frac{L_X^{\rm mini}}{\rm units}\right]', 'min': 40, 'max': 41},
    'M_TURN'            : {'tex_name' : r'\log_{10}\left[\frac{M_{\rm turn}}{{\rm M}_\odot}\right]', 'min': 7.6, 'max': 8.9, 'ticks' : []},
    'NU_X_THRESH'       : {'tex_name' : r'E_0~[\rm eV]', 'min': 300, 'max': 700, 'ticks' : []},
    'DM_LOG10_LIFETIME' : {'tex_name' : r'\log_{\rm 10}\left[\frac{\tau_{\chi}}{\rm s}\right]', 'min': 25.6, 'max': 26.4, 'ticks' : [], 'positive' : False},
    'DM_DECAY_RATE'     : {'tex_name' : r'\Gamma ~ [\rm s^{-1}]', 'positive' : True, 'val' : 0},
    'DM_LOG10_MASS'     : {'tex_name' : r'\log_{10}[\frac{m_{\chi}}{\rm eV}]', 'min': 3, 'max': 11,  'ticks' : []},
    'DM_FHEAT_APPROX_PARAM_LOG10_F0' : {'tex_name' : r'\log_{10}[f_0]', 'min': -2, 'max': 1,  'ticks' : []},
    'DM_FHEAT_APPROX_PARAM_A'  : {'tex_name' : r'a', 'min': -2, 'max': 1,  'ticks' : []},
    'DM_FHEAT_APPROX_PARAM_B'  : {'tex_name' : r'b', 'min': -2, 'max': 1,  'ticks' : []},
    'LOG10_XION_at_Z_HEAT_MAX' : {'tex_name' : r'\log_{10}[\bar x_e^{\rm init}]', 'min': -2, 'max': 1,  'ticks' : []},
    'LOG10_TK_at_Z_HEAT_MAX'   : {'tex_name' : r'\log_{10}[\frac{\bar T_K^{\rm init}}{\rm K}]', 'min': -2, 'max': 1,  'ticks' : []},
    'OMm'   : {'tex_name' : r'\Omega_{\rm m}', 'min': -2, 'max': 1,  'ticks' : []},
    'OMb'   : {'tex_name' : r'\Omega_{\rm b}', 'min': -2, 'max': 1,  'ticks' : []},
    'hlittle'   : {'tex_name' : r'h', 'min': -2, 'max': 1,  'ticks' : []},
    'SIGMA_8'   : {'tex_name' : r'\sigma_8', 'min': -2, 'max': 1,  'ticks' : []},
    'POWER_INDEX'   : {'tex_name' : r'n_{\rm s}', 'min': -2, 'max': 1,  'ticks' : []},
    'INVERSE_M_WDM'   : {'tex_name' : r'1/m_{\rm WDM}', 'min': -2, 'max': 1,  'ticks' : []},
    }


def make_triangle_plot(covariance, fiducial_params, **kwargs) : 

    #####################################
    ## Choose the data we want to look at
    
    if not isinstance(fiducial_params, list):
        fiducial_params[fiducial_params]

    if not isinstance(covariance, list):
        all_name_params = list(covariance.keys())
        covariance  = [covariance]  
    else:
        # Define all name_params
        all_name_params = []
        for name in [list(cov.keys()) for cov in covariance]:
            all_name_params += name
        all_name_params = list(set(all_name_params))  

    params_to_plot = kwargs.get('params_to_plot', all_name_params) 

    # add the fiducial values in the infos
    val_min = {}
    val_max = {}
    
    _default_info  = {'tex_name' : r'\theta', 'min' : None, 'max' : None, 'ticks' : [], 'positive' : False, 'val' : None}

    color = plt.rcParams['axes.prop_cycle'].by_key()['color']     
    color = kwargs.get('color', color)
    alpha = kwargs.get('alpha', [0.8]*len(covariance))

    if not isinstance(color, list):
        color = [color]

    for i_cov, cov in enumerate(covariance):
        for i_name in covariance[i_cov].keys():

            val = fiducial_params[i_cov][i_name]                   

            _param_info  = _PARAMS_PLOT.get(i_name, _default_info)
            if _param_info.get('val', None) is not None:
                val = _param_info.get('val', None)

            _sigma = np.sqrt(cov[i_name][i_name])
    
            val_min[i_name] = np.min([val - 3.5*_sigma, val_min.get(i_name, val)])
            val_max[i_name] = np.max([val + 3.5*_sigma, val_max.get(i_name, val)])

            if _param_info.get('positive', False) is True:
                val_min[i_name] = np.max([val_min[i_name], 0])
            
    
    fig, axs = prepare_triangle_plot(params_to_plot, val_min, val_max)
    for i_cov, cov in enumerate(covariance):
        fill_triangle_plot(cov, fiducial_params[i_cov], axs, color=color[i_cov], alpha = alpha[i_cov])

    axs[params_to_plot[0]][params_to_plot[0]].set_yticks([]) 
 
    return fig


def prepare_triangle_plot(params_to_plot, val_min, val_max):
    
    ngrid = len(params_to_plot)
    
    #####################################
    ##  Prepare the triangle plot
    fig = plt.figure(constrained_layout=False, figsize=(1.2*ngrid, 1.2*ngrid))
    fig.subplots_adjust(wspace=0.05, hspace=0.05)

    gs = GridSpec(ngrid, ngrid, figure=fig)
    axs = {param : {} for param in params_to_plot}
    
    for j, j_name in enumerate(params_to_plot) : 
        for i, i_name in enumerate(params_to_plot): 
            
            if i < j+1:

                ## Here i represents the x axis while j goes along the y axis
                axs[j_name][i_name] = fig.add_subplot(gs[j:j+1, i:i+1])

                ## Information concertning this case
                _default_info  = {'tex_name' : r'\theta', 'min' : None, 'max' : None, 'ticks' : [], 'positive' : False, 'val' : None}
                
                _param_info_x  = _PARAMS_PLOT.get(i_name, _default_info)
                _param_info_y  = _PARAMS_PLOT.get(j_name, _default_info)

                x_min = val_min[i_name]
                x_max = val_max[i_name]
                y_min = val_min[j_name]
                y_max = val_max[j_name]

                axs[j_name][i_name].set_xlim([x_min, x_max])

                if i != j :     
                    axs[j_name][i_name].set_ylim([y_min, y_max])

                if i == j :
                    axs[i_name][i_name].set_ylim([0, 1.2])


                ##############
                # Setting the plot

                # we remove the ticks if necessary for some parts of this case
                if j < ngrid -1 :
                    axs[j_name][i_name].xaxis.set_ticklabels([])
                if i > 0 : 
                    axs[j_name][i_name].yaxis.set_ticklabels([])

                if j == ngrid -1 :
                    axs[j_name][i_name].set_xlabel(r'${}$'.format(_param_info_x.get('tex_name', r'$\theta$')))
                    x_ticks = _param_info_x.get('ticks', [])
                    if x_ticks != [] :
                        axs[j_name][i_name].set_xticks(_param_info_x.get('ticks', []))
                    axs[j_name][i_name].tick_params(axis='x', labelsize=8)
                    for tick in axs[j_name][i_name].get_xticklabels():
                        tick.set_rotation(55)

                if i == 0 and j > 0:
                    axs[j_name][i_name].set_ylabel(r'${}$'.format(_param_info_y.get('tex_name', r'$\theta$')))
                    y_ticks = _param_info_y.get('ticks', [])
                    if y_ticks != [] :
                        axs[j_name][i_name].set_yticks(_param_info_y.get('ticks', []))
                    axs[j_name][i_name].tick_params(axis='y', labelsize=8)
                    for tick in axs[j_name][i_name].get_yticklabels():
                        tick.set_rotation(55)
           
    return fig, axs


def fill_triangle_plot(covariance, fiducial_params, axs, **kwargs):
    
    color = kwargs.get('color', None)
    alpha = kwargs.get('alpha', 0.8)

    params_to_plot = list(axs.keys())
    params_cov     = list(covariance.keys())

    params = []
    for param_p in params_to_plot:
        if param_p in params_cov:
            params.append(param_p)


    for j, j_name in enumerate(params) : 
        for i, i_name in enumerate(params): 

            if i < j+1:

                ##############
                val_x = fiducial_params[i_name]  
                val_y = fiducial_params[j_name]  

                ## Make the plots now
                if i != j : 

                    # Countour plot for the scatter
                    sub_cov = np.zeros((2, 2))
                    sub_cov[0, 0] = covariance[i_name][i_name]
                    sub_cov[0, 1] = covariance[i_name][j_name]
                    sub_cov[1, 0] = covariance[j_name][i_name]
                    sub_cov[1, 1] = covariance[j_name][j_name]

                    #ellipse_x, ellipse_y = ellipse_from_covariance(sub_cov, [val_x, val_y])
                    #axs[j_name][i_name].plot(ellipse_x, ellipse_y, linewidth=0.5, color='blue')

                    confidence_ellipse(sub_cov, val_x, val_y, axs[j_name][i_name],  n_std=2, facecolor=color, alpha=alpha/2.)
                    confidence_ellipse(sub_cov, val_x, val_y, axs[j_name][i_name],  n_std=1, facecolor=color, alpha=alpha)

                if i == j :

                    axs[i_name][i_name].set_ylim([0, 1.2])
                    axs[i_name][i_name].set_title(r'${}$'.format(val_x) + f'\n' + r'$\pm{:.3}$'.format(np.sqrt(covariance[i_name][i_name])), fontsize=10)
        
                    # Plot the gaussian approximation in that panel
                    sigma     = np.sqrt(covariance[i_name][i_name])
                    val_arr   = np.linspace(val_x-5*sigma, val_x+5*sigma, 100)
                    gaussian  = np.exp(-(val_arr - val_x)**2/2./sigma**2)

                    axs[i_name][i_name].plot(val_arr, gaussian, color=color)




def plot_func_vs_z_and_k(z, k, func, std = None, istd  : float = 0, **kwargs) :

    """ 
        Function that plots the power spectra with the sensitivity bounds from extract_noise_from_fiducial()
        We can plot on top more power spectra for comparison

        Params
        ------
        k : 1D array of floats
            modes 
        z : 1D array of floats
            redshifts
        func : (list of) 2D arrays of floats
            function(s) to plot in terms of the redshift and modes
        std : 1D array of floats
            standard deviation associated to func (or func[istd])
        istd : float, optional
            index of the func array where to attach the standard deviation
    """
    n_lines = int(len(z) / 5)

    if len(z) / 5 > n_lines:
        n_lines = n_lines + 1
    
    fig = plt.figure(constrained_layout=False, figsize=(10, n_lines*2))
    #fig.subplots_adjust()


    gs = GridSpec(n_lines, 5, figure=fig, wspace=0.5, hspace = 0.5)
    axs = [[None for j in range(0, 5)] for i in range(0, n_lines)]

    if not isinstance(func[0][0], (list, np.ndarray)) :
        func = [func]


    if len(func) > 1:
        
        cmap = mplt_cm.get_cmap('Spectral')
        a_lin = (0.99-0.2)/(len(func)-1) if len(func) > 1 else 1
        b_lin = 0.2 if len(func) > 1 else 0.5

        _default_color_list     = [cmap(i) for i in np.arange(0, len(k))*a_lin + b_lin]
        _default_linestyle_list = ['-' for i in np.arange(0, len(k))]
    
    else:

        _default_color_list     = ['b']
        _default_linestyle_list = ['-']


    
    color_list     = kwargs.get('color', _default_color_list)
    linestyle_list = kwargs.get('linestyle', _default_linestyle_list)
    ylabel         = kwargs.get('ylabel', None)
    title          = kwargs.get('title', None)
    marker         = kwargs.get('marker', None)
    markersize     = kwargs.get('markersize', None)
    
    no_line = Line2D([],[],color='k',linestyle='-',linewidth=0, alpha = 0) 

    # for the color of the fiducial to black
    color_list[istd] = 'k'
  
    iz = 0
    for i in range(0, n_lines):
        for j in range(0, 5):
            
            if iz < len(z): 

                axs[i][j] = fig.add_subplot(gs[i:i+1, j:j+1])

                # Plot the power spectrum at every redshift
                for jf, f in enumerate(func) : 
                    if marker is None : 
                        # By default we plot steps
                        axs[i][j].step(k, f[iz], where='mid', alpha = 1, color=color_list[jf], 
                                        linestyle = linestyle_list[jf], linewidth=0.5)
                    else: 
                        # Otherwise we use markers
                        axs[i][j].plot(k, f[iz], alpha = 1, color=color_list[jf], linestyle = linestyle_list[jf], 
                                        marker=marker, markersize=markersize, markerfacecolor=color_list[jf], 
                                        linewidth=0.8)
                    
                

                # Plot the standard deviation bars if standard deviation is given
                if std is not None : 
                    axs[i][j].fill_between(k, func[istd][iz] - std[iz], func[istd][iz] + std[iz], step='mid', alpha = 0.2, color=color_list[istd])
                
                
                xlim = kwargs.get('xlim', None)
                ylim = kwargs.get('ylim', None)

                if xlim is not None: 
                    axs[i][j].set_xlim(xlim)
                if ylim is not None: 
                    axs[i][j].set_ylim(ylim)

                xlog = kwargs.get('xlog', False)
                ylog = kwargs.get('ylog', False)

                if xlog is True:
                    axs[i][j].set_xscale('log')
                if ylog is True:
                    axs[i][j].set_yscale('log')
                

                axs[i][j].legend([no_line], [r'$\rm z = {0:.1f}$'.format(z[iz])], loc='lower right', handlelength=0, handletextpad=0, 
                                    bbox_to_anchor=(0.98, 1), fontsize=8, frameon=False,
                                    borderpad = 0, borderaxespad = 0, framealpha=0)

            iz = iz+1

    if ylabel is not None:
        axs[int(n_lines/2)-1][0].set_ylabel('{}'.format(ylabel))

    axs[n_lines-1][0].set_xlabel(r'$k ~{\rm [Mpc^{-1}]}$')


    if title is not None: 
        fig.suptitle('{}'.format(title), fontsize=14)


    return fig


def prepare_plot(**kwargs) :

    """ 
        Function that plots simple functions of one variable on the same graph
    """
    
    fig = plt.figure(figsize = (5,4))
    ax = fig.gca()

    xlim   = kwargs.get('xlim', None)
    ylim   = kwargs.get('ylim', None)
    xlog   = kwargs.get('xlog', False)
    ylog   = kwargs.get('ylog', False)
    xlabel = kwargs.get('xlabel', r'$x$')
    ylabel = kwargs.get('ylabel', r'$y$')

    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)
    ax.set_xlabel('{}'.format(xlabel))
    ax.set_ylabel('{}'.format(ylabel))

    if xlog is True:
        ax.set_xscale('log')
    if ylog is True:
        ax.set_yscale('log')

    return fig, ax



def prepare_2subplots(gridspec_kw = None, **kwargs) :

    """ 
        Function that plots simple functions of one variable on the same graph
    """
    
   

    xlim     = kwargs.get('xlim', None)
    ylim_1   = kwargs.get('ylim_1', None)
    ylim_2   = kwargs.get('ylim_2', None)
    xlog     = kwargs.get('xlog', False)
    ylog_1   = kwargs.get('ylog_1', False)
    ylog_2   = kwargs.get('ylog_2', False)
    xlabel   = kwargs.get('xlabel', r'$x$')
    ylabel_1 = kwargs.get('ylabel_1', r'$y$')
    ylabel_2 = kwargs.get('ylabel_2', r'$y$')
    figsize  = kwargs.get('figsize', (5, 4))

    fig, (ax1, ax2) = plt.subplots(2, 1, gridspec_kw=gridspec_kw, sharex=True, figsize=figsize, facecolor="White")

    if xlim is not None:
        ax1.set_xlim(xlim)
    if ylim_1 is not None:
        ax1.set_ylim(ylim_1)
    if ylim_2 is not None:
        ax2.set_ylim(ylim_2)

    ax2.set_xlabel('{}'.format(xlabel))
    ax1.set_ylabel('{}'.format(ylabel_1))
    ax2.set_ylabel('{}'.format(ylabel_2))

    if xlog is True:
        ax1.set_xscale('log')
    if ylog_1 is True:
        ax1.set_yscale('log')
    if ylog_2 is True:
        ax2.set_yscale('log')

    return fig, (ax1, ax2)




def plot_func(x, func, **kwargs) :

    """ 
        Function that plots simple functions of one variable on the same graph

        Params
        ------
        x : 1D array of floats
            redshifts
        func : (list of) 1D arrays of floats
            function(s) to plot in terms of x
    """

        
    fig, ax = prepare_plot(**kwargs)
    rax     = kwargs.get('rax', False)
    color   = kwargs.get('color', plt.rcParams['axes.prop_cycle'].by_key()['color'])
    
    if isinstance(func[0], (int, float, np.float64, np.float32)) : 
        func = [func]

    if isinstance(x[0], (int, float, np.float64, np.float32)) : 
        x = [x]*len(func)

    for ifunc, f in enumerate(func):
        ax.plot(x[ifunc], f, color=color[ifunc])

    if rax is False:
        return fig
    else:
        return fig, ax



def display_matrix(matrix, names = None):

    if names is not None:
        print('            ', end='')
        for name in names:
            print(name.ljust(10)[:10] + ' | ', end = '')

    print('')
    
    for i in range(0, len(matrix)):
        if names is not None:
            print(names[i].ljust(10)[:10] + ' | ', end='')
        for j in range(0, len(matrix)):
            add = ''
            if matrix[i][j] > 0:
                add = ' '
            print(add + "{:.1e}".format(matrix[i][j]) + '  |  ', end='')
        print('')




def load_uv_luminosity_functions(data_set = 'Bouwens21'):

    file_name = os.path.dirname(os.path.abspath(__file__)) + '/_data/' + data_set + '.npz'

    with open(file_name, 'rb') as file:  
        data = np.load(file, allow_pickle = True)   
       
        z_uv       = data['redshifts']
        m_uv       = data['M_uv']
        l_uv       = data['l_uv']
        sigma_l_uv = data['sigma_l_uv']


    return z_uv, m_uv, l_uv, sigma_l_uv




##################################################################################
# This file is part of 21cmCAST.
#
# Copyright (c) 2023, Gaétan Facchinetti
#
# 21cmCAST is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. 21cmCAST is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with 21cmCAST. 
# If not, see <https://www.gnu.org/licenses/>.
#
# -------------------------------------------------------------------------------
#
# Parts of this code have been copied and modified 
# from https://github.com/charlottenosam/21cmfish
# 
# - MIT License
# -
# - Copyright (c) 2019, Charlotte Mason
# - 
# - Permission is hereby granted, free of charge, to any person obtaining a copy
# - of this software and associated documentation files (the "Software"), to deal
# - in the Software without restriction, including without limitation the rights
# - to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# - copies of the Software, and to permit persons to whom the Software is
# - furnished to do so, subject to the following conditions:
# - 
# - The above copyright notice and this permission notice shall be included in all
# - copies or substantial portions of the Software.
# - 
# - THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# - IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# - FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# - AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# - LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# - OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# - SOFTWARE.
##################################################################################

import numpy as np
import powerbox.tools as pb_tools

from astropy.cosmology import Planck18 as cosmo
from astropy.cosmology import z_at_value

from astropy import units
from astropy import constants

from . import tools as tls



def deltak_zB(z, B) :
    
    # Definition of the 21 cm frequency (same as in 21cmSense)
    f21 = 1420.40575177 * units.MHz

    return 2*np.pi * f21 * cosmo.H(z) / constants.c / (1+z)**2 / B * 1000 * units.m / units.km


def generate_k_bins(z, k_min, k_max, B, logk):
    
    dk = deltak_zB(z, B) 
    _k_min = dk

    if _k_min < k_min :
        _k_min = k_min

    if logk is False:
        return np.arange(_k_min.value, k_max.value, dk.value) * k_min.unit
    else:
        ValueError("logarithmic k-bins not implemented yet")


def define_grid_modes_redshifts(z_min: float, B: float, k_min = 0.1 / units.Mpc, k_max = 1 / units.Mpc, z_max: float = 19, logk=False) : 
    """
    ## Defines a grid of modes and redshift on which to define the noise and on which the Fisher matrix will be evaluated
    
    Params:
    -------
    z_min : float
        Minimal redshift on the grid
    B     : float
        Bandwidth of the instrument

    """

    # Definition of the 21 cm frequency (same as in 21cmSense)
    f21 = 1420.40575177 * units.MHz

    def generate_z_bins(z_min, z_max, B):
        fmax = f21/(1+z_min)
        fmin = f21/(1+z_max)
        f_bins    = np.arange(fmax.value, fmin.value, -B.value) * f21.unit
        f_centers = f_bins[:-1] - B/2
        z_bins    = (f21/f_bins).value - 1
        z_centers = (f21/f_centers).value -1
        return z_bins, z_centers

    # Get the redshift bin edges and centers
    z_bins, z_centers = generate_z_bins(z_min, z_max, B)
    
    # Get the k-bins edges
    k_bins = generate_k_bins(z_min, k_min, k_max, B, logk)

    return z_bins, z_centers, k_bins



def get_k_min_max(lightcone, n_chunks=24):
    """
    Get the minimum and maximum k in 1/Mpc to calculate powerspectra for
    given size of box and number of chunks
    """

    BOX_LEN = lightcone.user_params.pystruct['BOX_LEN']
    HII_DIM = lightcone.user_params.pystruct['HII_DIM']

    k_fundamental = 2*np.pi/BOX_LEN*max(1,len(lightcone.lightcone_distances)/n_chunks/HII_DIM) #either kpar or kperp sets the min
    k_max         = k_fundamental * HII_DIM
    Nk            = np.floor(HII_DIM/1).astype(int)
    return k_fundamental, k_max, Nk


def compute_power(box,
                   length,
                   n_psbins,
                   log_bins=True,
                   k_min=None,
                   k_max=None,
                   ignore_kperp_zero=True,
                   ignore_kpar_zero=False,
                   ignore_k_zero=False):
    """
    Calculate power spectrum for a redshift chunk
    TODO
    Parameters
    ----------
    box :
        lightcone brightness_temp chunk
    length :
        TODO
    n_psbins : int
        number of k bins
    Returns
    ----------
        k : 1/Mpc
        delta : mK^2
        err_delta : mK^2
    """
    # Determine the weighting function required from ignoring k's.
    k_weights = np.ones(box.shape, dtype=int)
    n0 = k_weights.shape[0]
    n1 = k_weights.shape[-1]

    if ignore_kperp_zero:
        k_weights[n0 // 2, n0 // 2, :] = 0
    if ignore_kpar_zero:
        k_weights[:, :, n1 // 2] = 0
    if ignore_k_zero:
        k_weights[n0 // 2, n0 // 2, n1 // 2] = 0

    # Define k bins
    if (k_min is None and k_max is None) or n_psbins is None:
        bins = n_psbins
    elif isinstance(n_psbins, int): 
        # In the case where n_psbins is just an integer
        if log_bins:
            bins = np.logspace(np.log10(k_min), np.log10(k_max), n_psbins)
        else:
            bins = np.linspace(k_min, k_max, n_psbins)
    else:
        bins = n_psbins

    res = pb_tools.get_power(
        box,
        boxlength=length,
        bins=bins,
        bin_ave=False,
        get_variance=True,
        log_bins=log_bins,
        k_weights=k_weights,
    )

    res = list(res)
    k = res[1]
    if log_bins:
        k = np.exp((np.log(k[1:]) + np.log(k[:-1])) / 2)
    else:
        k = (k[1:] + k[:-1]) / 2

    res[1] = k

    return res


def compute_powerspectra_1D(lightcone, nchunks=15,
                    chunk_indices=None,
                    n_psbins=None,
                    k_min=0.1,
                    k_max=1.0,
                    logk=True,
                    ignore_kperp_zero=True,
                    ignore_kpar_zero=False,
                    ignore_k_zero=False,
                    remove_nans=False,
                    vb=False):

    """
    Make power spectra for given number of equally spaced redshift chunks OR list of redshift chunk lightcone indices
    Output:
        k : 1/Mpc
        delta : mK^2
        err_delta : mK^2
    TODO this isn't using k_min, k_max...
    """
    data = []

    # Create lightcone redshift chunks
    # If chunk indices not given, divide lightcone into nchunks equally spaced redshift chunks
    if chunk_indices is None:
        #print(lightcone.n_slices, lightcone.brightness_temp.shape, round(lightcone.n_slices / nchunks))
        chunk_indices = list(range(0, lightcone.n_slices, int(lightcone.n_slices / nchunks)))
    else:
        nchunks = len(chunk_indices) - 1

    chunk_redshift = np.zeros(nchunks)
    z_centers      = np.zeros(nchunks)

    lc_redshifts = lightcone.lightcone_redshifts
    lc_distances = lightcone.lightcone_distances

    # Calculate PS in each redshift chunk
    for i in range(nchunks):
        
        if vb:
            print(f'Chunk {i}/{nchunks}...')
        
        start    = chunk_indices[i]
        end      = chunk_indices[i + 1]
        chunklen = (end - start) * lightcone.cell_size

        chunk_redshift[i] = np.median(lc_redshifts[start:end])


        #####

        index_center = int(lightcone.n_slices / nchunks)
           
        if index_center % 2 == 0:
            dist_center = 0.5 * ( lc_distances[start + int(0.5 * index_center)] + lc_distances[start + int(0.5 * index_center) - 1])
        else:
            dist_center = lc_distances[start + int(0.5 * index_center)]
            
        z_centers[i] = z_at_value(lightcone.cosmo_params.cosmo.comoving_distance, dist_center * units.Mpc)

        #####


        if chunklen == 0:
            print(f'Chunk size = 0 for z = {lc_redshifts[start]}-{lc_redshifts[end]}')
        else:
            power, k, variance = compute_power(
                    lightcone.brightness_temp[:, :, start:end],
                    (lightcone.user_params.BOX_LEN, lightcone.user_params.BOX_LEN, chunklen),
                    n_psbins,
                    log_bins=logk,
                    k_min=k_min,
                    k_max=k_max,
                    ignore_kperp_zero=ignore_kperp_zero,
                    ignore_kpar_zero=ignore_kpar_zero,
                    ignore_k_zero=ignore_k_zero,)

            if remove_nans:
                power, k, variance = power[~np.isnan(power)], k[~np.isnan(power)], variance[~np.isnan(power)]
            else:
                variance[np.isnan(power)] = np.inf

            data.append({"k": k, "delta": power * k ** 3 / (2 * np.pi ** 2), "err_delta": np.sqrt(variance) * k ** 3 / (2 * np.pi ** 2)})

    return z_centers, data


def export_powerspectra_1D_vs_k(path, z_centers, power_spectra, clean_existing_dir: bool = True) -> None:
    """ Export the power_spectra obtained from compute_power_spectra_1D """
    
    # Make the directories associated to path
    tls.make_directory(path + "/power_spectra_vs_k", clean_existing_dir)

    for iz, z in enumerate(z_centers):
        save_path_ps = path + '/power_spectra_vs_k/ps_z_' + "{0:.1f}".format(z) + '.txt' 

        with open(save_path_ps, 'w') as f:
            print("# Power spectrum at redshit z = " + str(z), file=f)
            print("# k [Mpc^{-1}] | Delta_{21}^2 [mK^2] | err_Delta [mK^2]", file=f)
            for ik, k in enumerate(power_spectra[iz]['k']): 
                print(str(k) + "\t" +  str(power_spectra[iz]['delta'][ik]) + "\t" +  str(power_spectra[iz]['err_delta'][ik]), file=f)

    save_path_redshifts = path + '/power_spectra_vs_k/redshift_chunks.txt'
    with open(save_path_redshifts, 'w') as f:
        print("# Redshift chunks at which the power spectrum is computed", file=f)
        for z in z_centers: 
            print(z, file=f)



def export_powerspectra_1D_vs_z(path, z_centers, power_spectra, clean_existing_dir:bool = True) -> None :
    """ Export the power_spectra obtained from compute_power_spectra_1D """
    
    # Make the directories associated to path
    tls.make_directory(path + "/power_spectra_vs_z", clean_existing_dir)

    for ik, k in enumerate(power_spectra[0]['k']):
        save_path_ps = path + '/power_spectra_vs_z/ps_k_' + "{0:.6f}".format(k) + '.txt' 

        with open(save_path_ps, 'w') as f:
            print("# Power spectrum at mode k = " + str(k), file=f)
            print("# z [Mpc^{-1}] | Delta_{21}^2 [mK^2] | err_Delta [mK^2]", file=f)
            for iz, z in enumerate(z_centers): 
                print(str(z) + "\t" +  str(power_spectra[iz]['delta'][ik]) + "\t" +  str(power_spectra[iz]['err_delta'][ik]), file=f)


    save_path_k = path + '/power_spectra_vs_z/k_chunks.txt'
    with open(save_path_k, 'w') as f:
        print("# k values at which the power spectrum is computed [Mpc^{-1}]", file=f)
        for k in power_spectra[0]['k']: 
            print(k, file=f)




##################################################################################
# This file is part of 21cmCAST.
#
# Copyright (c) 2022, Gaétan Facchinetti
#
# 21cmCAST is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. 21cmCAST is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with 21cmCAST. 
# If not, see <https://www.gnu.org/licenses/>.
##################################################################################


import numpy  as np
from astropy import units
from scipy   import interpolate

PY21CMSENSE = True

## This part of the code rely on 21cmSense however as many of
## 21cmCAST abilities do not rely on 21cmSense we allow using 
## 21cmCAST without 21cmSense functionalities


# create a simple observations class
class ObservationSet:
    
    def __init__(self, name, *, observations = None):
        self._observations = observations
        self._name = name

    @property
    def observations(self):
        return self._observations
    
    @property 
    def name(self):
        return self._name


try :

    import py21cmsense as p21s
    

    def default_HERA_observatories(z_arr:np.ndarray):

        ## Define the layout for the 
        hera_layout = p21s.antpos.hera(
            hex_num = 11,             # number of antennas along a side
            separation= 14 * units.m,  # separation between antennas (in metres)
            row_separation=12.12 * units.m        # separation between rows
        )

        hera_observatories = []
        
        for z in z_arr:
            beam = p21s.beam.GaussianBeam(
               frequency = 1420.40575177 * units.MHz / (1+z),  # just a reference frequency
              dish_size = 14 * units.m
            )

            hera_observatories.append(p21s.Observatory(
                antpos = hera_layout,
                beam = beam,
                latitude = 0.536189 * units.radian,
                Trcv = 100 * units.K
            ))

        return hera_observatories


    def default_observation_set(z_arr:np.ndarray):
  
        default_observatories = default_HERA_observatories(z_arr)

        observations = []
        for observatory in default_observatories:

                observations.append(p21s.Observation(
                    observatory   = observatory,
                    n_channels    = 80, 
                    bandwidth     = 8 * units.MHz,
                    time_per_day  = 6 * units.hour,   # Number of hours of observation per day
                    n_days        = 166.6676,         # Number of days of observation
                ))

        return ObservationSet("default_HERA", observations = observations)
        

    class ComputedModel(p21s.theory.TheoryModel):

        use_littleh = False

        def __init__(self, k_21:np.ndarray, delta_21_sqr:np.ndarray) -> None:
            
            self.k = k_21
            self.delta_sqr = delta_21_sqr

            # interpolation function for this redshift
            self.interp = interpolate.interp1d(k_21, delta_21_sqr / (units.mK)**2)

        # Note that here z is a dummy variable
        def delta_squared(self, z: float, k: np.ndarray) -> units.Quantity[units.mK**2]:
            return self.interp((k * units.Mpc)) * (units.mK)**2



    def extract_noise_from_fiducial(k_in, dsqr_in, k_out, observation) :
        """
        Give the noise associated to power spectra delta_arr

        Params:
        -------
        k       : list of floats 
            array of modes k in [Mpc^{-1}]
        dsqr    : list of floats 
            Power spectrum in [mK^2] ordered with the modes k in 
        observation : Observation object (c.f. 21cmSense)

        Returns:
        --------
        k_sens       : list of list of floats
        std_sens     : the standard deviation of the sensitivity [mK]

        """

        #sensitivity       = p21s.PowerSpectrum(observation=observation, k_21 = k_in / units.Mpc, 
        #                                        delta_21 = dsqr_in * (units.mK**2), 
        #                                        foreground_model='moderate') 
        
        ## implementation compatible with the new version of 21cmSense
        theory_model = ComputedModel(k_21 = k_in / units.Mpc, delta_21_sqr = dsqr_in * (units.mK**2))
        sensitivity  = p21s.PowerSpectrum(observation = observation, theory_model = theory_model, foreground_model = 'moderate')
        
        k_sens            = sensitivity.k1d.value * sensitivity.cosmo.h
        std_21cmSense     = sensitivity.calculate_sensitivity_1d(thermal = True, sample = True).value

        std = interpolate.interp1d(k_sens, std_21cmSense, bounds_error=False, fill_value=np.inf)(k_out)

        return std

except ImportError:

    PY21CMSENSE = False
    
    def default_HERA_observatories(z_arr):
        return None
    
    def default_observation_set(z_arr):
        return ObservationSet("default_HERA", None)

    def extract_noise_from_fiducial(k_in, dsqr_in, k_out, observation):
        return None